import datetime
import multiprocessing
import os
import sys
from typing import Self

import limexhub
import structlog
from asyncclick import progressbar
from joblib import Parallel, delayed

import polars as pl
import yfinance as yf
from ziplime.data.data_sources.asset_data_source import AssetDataSource
from ziplime.data.services.data_bundle_source import DataBundleSource


class YahooFinanceAssetDataSource(AssetDataSource):
    def __init__(self, maximum_threads: int | None = None):
        super().__init__()
        self._logger = structlog.get_logger(__name__)
        if maximum_threads is not None:
            self._maximum_threads = min(multiprocessing.cpu_count() * 2, maximum_threads)
        else:
            self._maximum_threads = multiprocessing.cpu_count() * 2

    async def get_assets(self, symbols: list[str], **kwargs) -> pl.DataFrame:
        assets = yf.Tickers(' '.join(symbols))
        return assets

    async def search_assets(self, query: str, **kwargs) -> pl.DataFrame:
        assets = yf.Lookup(query=query).all
        return assets

    async def get_constituents(self, index: str) -> pl.DataFrame:
        index = "SPY"
        ticker = yf.Ticker(index)
        funds_data = ticker.funds_data
        try:
            holdings = funds_data.equity_holdings
        except Exception as e:
            holdings = pl.DataFrame()

        return assets




"""
This module defines the interface and implementations of Pipeline domains.

A domain represents a set of labels for the arrays computed by a Pipeline.
Currently, this means that a domain defines two things:

1. A calendar defining the dates to which the pipeline's inputs and outputs
   should be aligned. The calendar is represented concretely by a pandas
   DatetimeIndex.

2. The set of assets that the pipeline should compute over. Right now, the only
   supported way of representing this set is with a two-character country code
   describing the country of assets over which the pipeline should compute. In
   the future, we expect to expand this functionality to include more general
   concepts.
"""

import datetime
from textwrap import dedent

import numpy as np
import pandas as pd
import pytz

from ziplime.utils.calendar_utils import get_calendar

from ziplime.country import CountryCode
from ziplime.utils.formatting import bulleted_list
from exchange_calendars.utils.pandas_utils import days_at_time


class Domain:
    """Domain interface."""

    def sessions(self):
        """Get all trading sessions for the calendar of this domain.

        This determines the row labels of Pipeline outputs for pipelines run on
        this domain.

        Returns
        -------
        sessions : pd.DatetimeIndex
            An array of all session labels for this domain.
        """

    @property
    def country_code(self):
        """The country code for this domain.

        Returns
        -------
        code : str
            The two-character country iso3166 country code for this domain.
        """

    def data_query_cutoff_for_sessions(self, sessions):
        """Compute the data query cutoff time for the given sessions.

        Parameters
        ----------
        sessions : pd.DatetimeIndex
            The sessions to get the data query cutoff times for. This index
            will contain all midnight UTC values.

        Returns
        -------
        data_query_cutoff : pd.DatetimeIndex
            Timestamp of the last minute for which data should be considered
            "available" on each session.
        """

    def roll_forward(self, dt):
        """Given a date, align it to the calendar of the pipeline's domain.

        Parameters
        ----------
        dt : pd.Timestamp

        Returns
        -------
        pd.Timestamp
        """
        dt = pd.Timestamp(dt)
        trading_days = self.sessions()
        try:
            return trading_days[trading_days.searchsorted(dt)]
        except IndexError as exc:
            raise ValueError(
                f"Date {dt.date()} was past the last session for domain {self}. "
                f"The last session for this domain is {trading_days[-1].date()}."
            ) from exc


Domain.__doc__ = """A domain represents a set of labels for the arrays computed by a Pipeline.

A domain defines two things:

1. A calendar defining the dates to which the pipeline's inputs and outputs
   should be aligned. The calendar is represented concretely by a pandas
   DatetimeIndex.

2. The set of assets that the pipeline should compute over. Right now, the only
   supported way of representing this set is with a two-character country code
   describing the country of assets over which the pipeline should compute. In
   the future, we expect to expand this functionality to include more general
   concepts.
"""
Domain.__name__ = "Domain"
Domain.__qualname__ = "ziplime.pipeline.domain.Domain"


class GenericDomain(Domain):
    """Special singleton class used to represent generic DataSets and Columns."""

    def sessions(self):
        raise NotImplementedError("Can't get sessions for generic domain.")

    @property
    def country_code(self):
        raise NotImplementedError("Can't get country code for generic domain.")

    def data_query_cutoff_for_sessions(self, sessions):
        raise NotImplementedError(
            "Can't compute data query cutoff times for generic domain.",
        )

    def __repr__(self):
        return "GENERIC"


GENERIC = GenericDomain()


class EquityCalendarDomain(Domain):
    """An equity domain whose sessions are defined by a named TradingCalendar.

    Parameters
    ----------
    country_code : str
        ISO-3166 two-letter country code of the domain
    calendar_name : str
        Name of the calendar, to be looked by by trading_calendar.get_calendar.
    data_query_offset : np.timedelta64
         The offset from market open when data should no longer be considered
         available for a session. For example, a ``data_query_offset`` of
         ``-np.timedelta64(45, 'm')`` means that the data must have
         been available at least 45 minutes prior to market open for it to
         appear in the pipeline input for the given session.
    """

    def __init__(
            self, country_code: str, calendar_name:str, data_query_offset=-np.timedelta64(45, "m")
    ):
        self._country_code = country_code
        self.calendar_name = calendar_name
        self._data_query_offset = (
            # add one minute because `open_time` is actually the open minute
            # label which is one minute _after_ market open...
                data_query_offset
                - np.timedelta64(1, "m")
        )
        if data_query_offset >= datetime.timedelta(0):
            raise ValueError(
                "data must be ready before market open (offset must be < 0)",
            )

    @property
    def country_code(self):
        return self._country_code

    #@lazyval
    @property
    def calendar(self):
        return get_calendar(self.calendar_name)

    def sessions(self):
        return self.calendar.sessions

    def data_query_cutoff_for_sessions(self, sessions):
        opens = self.calendar.first_minutes.reindex(sessions)
        missing_mask = pd.isnull(opens)
        if missing_mask.any():
            missing_days = sessions[missing_mask]
            raise ValueError(
                "cannot resolve data query time for sessions that are not on"
                f" the {self.calendar_name} calendar:\n{missing_days}"
            )

        return pd.DatetimeIndex(opens) + self._data_query_offset

    def __repr__(self):
        return "EquityCalendarDomain({!r}, {!r})".format(
            self.country_code,
            self.calendar_name,
        )


AR_EQUITIES = EquityCalendarDomain(CountryCode.ARGENTINA, "XBUE")
AT_EQUITIES = EquityCalendarDomain(CountryCode.AUSTRIA, "XWBO")
AU_EQUITIES = EquityCalendarDomain(CountryCode.AUSTRALIA, "XASX")
BE_EQUITIES = EquityCalendarDomain(CountryCode.BELGIUM, "XBRU")
BR_EQUITIES = EquityCalendarDomain(CountryCode.BRAZIL, "BVMF")
CA_EQUITIES = EquityCalendarDomain(CountryCode.CANADA, "XTSE")
CH_EQUITIES = EquityCalendarDomain(CountryCode.SWITZERLAND, "XSWX")
CL_EQUITIES = EquityCalendarDomain(CountryCode.CHILE, "XSGO")
CN_EQUITIES = EquityCalendarDomain(CountryCode.CHINA, "XSHG")
CO_EQUITIES = EquityCalendarDomain(CountryCode.COLOMBIA, "XBOG")
CZ_EQUITIES = EquityCalendarDomain(CountryCode.CZECH_REPUBLIC, "XPRA")
DE_EQUITIES = EquityCalendarDomain(CountryCode.GERMANY, "XFRA")
DK_EQUITIES = EquityCalendarDomain(CountryCode.DENMARK, "XCSE")
ES_EQUITIES = EquityCalendarDomain(CountryCode.SPAIN, "XMAD")
FI_EQUITIES = EquityCalendarDomain(CountryCode.FINLAND, "XHEL")
FR_EQUITIES = EquityCalendarDomain(CountryCode.FRANCE, "XPAR")
GB_EQUITIES = EquityCalendarDomain(CountryCode.UNITED_KINGDOM, "XLON")
GR_EQUITIES = EquityCalendarDomain(CountryCode.GREECE, "ASEX")
HK_EQUITIES = EquityCalendarDomain(CountryCode.HONG_KONG, "XHKG")
HU_EQUITIES = EquityCalendarDomain(CountryCode.HUNGARY, "XBUD")
ID_EQUITIES = EquityCalendarDomain(CountryCode.INDONESIA, "XIDX")
IE_EQUITIES = EquityCalendarDomain(CountryCode.IRELAND, "XDUB")
IN_EQUITIES = EquityCalendarDomain(CountryCode.INDIA, "XBOM")
IT_EQUITIES = EquityCalendarDomain(CountryCode.ITALY, "XMIL")
JP_EQUITIES = EquityCalendarDomain(CountryCode.JAPAN, "XTKS")
KR_EQUITIES = EquityCalendarDomain(CountryCode.SOUTH_KOREA, "XKRX")
MX_EQUITIES = EquityCalendarDomain(CountryCode.MEXICO, "XMEX")
MY_EQUITIES = EquityCalendarDomain(CountryCode.MALAYSIA, "XKLS")
NL_EQUITIES = EquityCalendarDomain(CountryCode.NETHERLANDS, "XAMS")
NO_EQUITIES = EquityCalendarDomain(CountryCode.NORWAY, "XOSL")
NZ_EQUITIES = EquityCalendarDomain(CountryCode.NEW_ZEALAND, "XNZE")
PE_EQUITIES = EquityCalendarDomain(CountryCode.PERU, "XLIM")
PH_EQUITIES = EquityCalendarDomain(CountryCode.PHILIPPINES, "XPHS")
PK_EQUITIES = EquityCalendarDomain(CountryCode.PAKISTAN, "XKAR")
PL_EQUITIES = EquityCalendarDomain(CountryCode.POLAND, "XWAR")
PT_EQUITIES = EquityCalendarDomain(CountryCode.PORTUGAL, "XLIS")
RU_EQUITIES = EquityCalendarDomain(CountryCode.RUSSIA, "XMOS")
SE_EQUITIES = EquityCalendarDomain(CountryCode.SWEDEN, "XSTO")
SG_EQUITIES = EquityCalendarDomain(CountryCode.SINGAPORE, "XSES")
TH_EQUITIES = EquityCalendarDomain(CountryCode.THAILAND, "XBKK")
TR_EQUITIES = EquityCalendarDomain(CountryCode.TURKEY, "XIST")
TW_EQUITIES = EquityCalendarDomain(CountryCode.TAIWAN, "XTAI")
US_EQUITIES = EquityCalendarDomain(CountryCode.UNITED_STATES, "XNYS")
ZA_EQUITIES = EquityCalendarDomain(CountryCode.SOUTH_AFRICA, "XJSE")

BUILT_IN_DOMAINS = [
    AR_EQUITIES,
    AT_EQUITIES,
    AU_EQUITIES,
    BE_EQUITIES,
    BR_EQUITIES,
    CA_EQUITIES,
    CH_EQUITIES,
    CL_EQUITIES,
    CN_EQUITIES,
    CO_EQUITIES,
    CZ_EQUITIES,
    DE_EQUITIES,
    DK_EQUITIES,
    ES_EQUITIES,
    FI_EQUITIES,
    FR_EQUITIES,
    GB_EQUITIES,
    GR_EQUITIES,
    HK_EQUITIES,
    HU_EQUITIES,
    ID_EQUITIES,
    IE_EQUITIES,
    IN_EQUITIES,
    IT_EQUITIES,
    JP_EQUITIES,
    KR_EQUITIES,
    MX_EQUITIES,
    MY_EQUITIES,
    NL_EQUITIES,
    NO_EQUITIES,
    NZ_EQUITIES,
    PE_EQUITIES,
    PH_EQUITIES,
    PK_EQUITIES,
    PL_EQUITIES,
    PT_EQUITIES,
    RU_EQUITIES,
    SE_EQUITIES,
    SG_EQUITIES,
    TH_EQUITIES,
    TR_EQUITIES,
    TW_EQUITIES,
    US_EQUITIES,
    ZA_EQUITIES,
]


def infer_domain(terms):
    """Infer the domain from a collection of terms.

    The algorithm for inferring domains is as follows:

    - If all input terms have a domain of GENERIC, the result is GENERIC.

    - If there is exactly one non-generic domain in the input terms, the result
      is that domain.

    - Otherwise, an AmbiguousDomain error is raised.

    Parameters
    ----------
    terms : iterable[ziplime.pipeline.Term]

    Returns
    -------
    inferred : Domain or None

    Raises
    ------
    AmbiguousDomain
        Raised if more than one concrete domain is present in the input terms.
    """
    domains = {t.domain for t in terms}
    num_domains = len(domains)

    if num_domains == 0:
        return GENERIC
    elif num_domains == 1:
        return domains.pop()
    elif num_domains == 2 and GENERIC in domains:
        domains.remove(GENERIC)
        return domains.pop()
    else:
        # Remove GENERIC if it's present before raising. Showing it to the user
        # is confusing because it doesn't contribute to the error.
        domains.discard(GENERIC)
        raise AmbiguousDomain(sorted(domains, key=repr))


# This would be better if we provided more context for which domains came from
# which terms.
class AmbiguousDomain(Exception):
    """Raised when we attempt to infer a domain from a collection of mixed terms."""

    _TEMPLATE = dedent(
        """\
        Found terms with conflicting domains:
        {domains}"""
    )

    def __init__(self, domains):
        self.domains = domains

    def __str__(self):
        return self._TEMPLATE.format(
            domains=bulleted_list(self.domains, indent=2),
        )


class EquitySessionDomain(Domain):
    """A domain built directly from an index of sessions.

    Mostly useful for testing.

    Parameters
    ----------
    sessions : pd.DatetimeIndex
        Sessions to use as output labels for pipelines run on this domain.
    country_code : str
        ISO 3166 country code of equities to be used with this domain.
    data_query_time : datetime.time, optional
        The time of day when data should no longer be considered available for
        a session.
    data_query_date_offset : int, optional
        The number of days to add to the session label before applying the
        ``data_query_time``. This can be used to express that the cutoff time
        for a session falls on a different calendar day from the session label.
    """

    def __init__(
            self,
            sessions: pd.DatetimeIndex,
            country_code: str,
            data_query_time: datetime.time | None = None,
            data_query_date_offset: int = 0,
    ):
        self._country_code = country_code
        self._sessions = sessions

        if data_query_time is None:
            data_query_time = datetime.time(0, 0, tzinfo=pytz.timezone("UTC"))

        if data_query_time.tzinfo is None:
            raise ValueError("data_query_time cannot be tz-naive")

        self._data_query_time = data_query_time
        self._data_query_date_offset = data_query_date_offset

    @property
    def country_code(self):
        return self._country_code

    def sessions(self):
        return self._sessions

    def data_query_cutoff_for_sessions(self, sessions):
        return days_at_time(
            sessions,
            self._data_query_time,
            self._data_query_time.tzinfo,
            self._data_query_date_offset,
        )



from enum import Enum


class PositionAction(Enum):
    """
    Enum representing position actions for derivatives trading.
    """
    OPEN = "OPEN"  # Open a new position or add to existing
    CLOSE = "CLOSE"  # Close an existing position
    NIL = "NIL"  # No specific position action (e.g., for spot trading)
    REDUCE = "REDUCE"  # Reduce size of an existing position
    FLIP = "FLIP"  # Change position from long to short or vice versa



import datetime
import math
from collections import OrderedDict

import numpy as np
import pandas as pd
import structlog

from ziplime.assets.entities.asset import Asset
from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.domain.account import Account
from ziplime.domain.portfolio import Portfolio
from ziplime.exchanges.exchange import Exchange
from ziplime.finance.commission import CommissionModel

from ziplime.finance.domain.order import Order
from ziplime.finance.domain.position_tracker import PositionTracker
from ziplime.finance.domain.transaction import Transaction


class Ledger:
    """The ledger tracks all orders and transactions as well as the current
    state of the portfolio and positions.

    Attributes
    ----------
    portfolio : ziplime.protocol.Portfolio
        The updated portfolio being managed.
    account : ziplime.protocol.Account
        The updated account being managed.
    position_tracker : PositionTracker
        The current set of positions.
    todays_returns : float
        The current day's returns. In minute emission mode, this is the partial
        day's returns. In daily emission mode, this is
        ``daily_returns[session]``.
    daily_returns_series : pd.Series
        The daily returns series. Days that have not yet finished will hold
        a value of ``np.nan``.
    daily_returns_array : np.ndarray
        The daily returns as an ndarray. Days that have not yet finished will
        hold a value of ``np.nan``.
    """

    def __init__(self, trading_sessions: pd.DatetimeIndex,
                 exchanges: dict[str, Exchange],
                 data_frequency: datetime.timedelta):
        if len(trading_sessions):
            start = trading_sessions[0]
        else:
            start = None
        # Have some fields of the portfolio changed? This should be accessed
        # through ``self._dirty_portfolio``
        self.__dirty_portfolio = False

        first_exchange = exchanges[list(exchanges.keys())[0]]
        self.default_exchange = first_exchange
        self._portfolio = Portfolio(start_date=start,
                                    starting_cash=self.default_exchange.get_start_cash_balance(),
                                    portfolio_value=self.default_exchange.get_start_cash_balance(),
                                    cash=self.default_exchange.get_current_cash_balance(),
                                    cash_flow=0.00,
                                    pnl=0.00,
                                    returns=0.00,
                                    positions_value=0.00,
                                    positions_exposure=0.00,
                                    )

        self.logger = structlog.get_logger(__name__)

        self.daily_returns_series = pd.Series(
            np.nan,
            index=trading_sessions,
        )
        # Get a view into the storage of the returns series. Metrics
        # can access this directly in minute mode for performance reasons.
        self.daily_returns_array = self.daily_returns_series.values

        self._previous_total_returns = 0

        # this is a component of the cache key for the account
        self._position_stats = None

        # Have some fields of the account changed?
        self._dirty_account = True
        # self._immutable_account =
        self._account = Account(
            settled_cash=0.00,
            accrued_interest=0.00,
            buying_power=math.inf,
            equity_with_loan=0.00,
            total_positions_value=0.00,
            total_positions_exposure=0.00,
            regt_equity=0.00,
            regt_margin=math.inf,
            initial_margin_requirement=0.00,
            maintenance_margin_requirement=0.00,
            available_funds=0.00,
            excess_liquidity=0.00,
            cushion=0.00,
            day_trades_remaining=math.inf,
            leverage=0.00,
            net_leverage=0.00,
            net_liquidation=0.00
        )

        # The exchange blotter can override some fields on the account. This is
        # way to tangled up at the moment but we aren't fixing it today.
        self._account_overrides = {}
        self._data_frequency = data_frequency

        self.position_tracker = PositionTracker(exchanges=exchanges, data_frequency=data_frequency)

        self._processed_transactions = {}

        self._orders_by_modified = {}
        self._orders_by_id = OrderedDict()

        # Keyed by asset, the previous last sale price of positions with
        # payouts on price differences, e.g. Futures.
        #
        # This dt is not the previous minute to the minute for which the
        # calculation is done, but the last sale price either before the period
        # start, or when the price at execution.
        self._payout_last_sale_prices = {}

    @property
    def todays_returns(self) -> float:
        # compute today's returns in returns space instead of portfolio-value
        # space to work even when we have capital changes
        return (self.portfolio.returns + 1) / (self._previous_total_returns + 1) - 1

    @property
    def _dirty_portfolio(self):
        return self.__dirty_portfolio

    @_dirty_portfolio.setter
    def _dirty_portfolio(self, value):
        if value:
            # marking the portfolio as dirty also marks the account as dirty
            self.__dirty_portfolio = self._dirty_account = value
        else:
            self.__dirty_portfolio = value

    def start_of_session(self, session_label):
        self._processed_transactions.clear()
        self._orders_by_modified.clear()
        self._orders_by_id.clear()

        # Save the previous day's total returns so that ``todays_returns``
        # produces returns since yesterday. This does not happen in
        # ``end_of_session`` because we want ``todays_returns`` to produce the
        # correct value in metric ``end_of_session`` handlers.
        self._previous_total_returns = self.portfolio.returns

    def end_of_bar(self, session_ix):
        # make daily_returns hold the partial returns, this saves many
        # metrics from doing a concat and copying all of the previous
        # returns
        if isinstance(self.daily_returns_array, np.ndarray):
            self.daily_returns_array[session_ix] = self.todays_returns
        elif isinstance(self.daily_returns_array, pd.Series):
            self.daily_returns_array.iloc[session_ix] = self.todays_returns
        else:
            raise ValueError("Unknown daily returns array type")

    def end_of_session(self, session_ix: int):
        # save the daily returns time-series
        self.daily_returns_series.iloc[session_ix] = self.todays_returns

    def sync_last_sale_prices(self, dt: datetime.datetime, handle_non_market_minutes: bool = False):
        self.position_tracker.sync_last_sale_prices(
            dt=dt,
            handle_non_market_minutes=handle_non_market_minutes,
            exchange_name=self.default_exchange.name
        )
        self._dirty_portfolio = True

    @staticmethod
    def _calculate_payout(multiplier: float, amount: float, old_price: float, price: float) -> float:
        return (price - old_price) * multiplier * amount

    def _cash_flow(self, amount: float) -> None:
        self._dirty_portfolio = True
        self._portfolio.cash_flow += amount
        self._portfolio.cash += amount

    def process_transaction(self, transaction: Transaction):
        """Add a transaction to ledger, updating the current state as needed.

        Parameters
        ----------
        transaction : Transaction
            The transaction to execute.
        """
        asset = transaction.asset
        if isinstance(asset, FuturesContract):
            try:
                old_price = self._payout_last_sale_prices[asset]
            except KeyError:
                self._payout_last_sale_prices[asset] = transaction.price
            else:
                position = self.position_tracker.positions[asset]
                amount = position.amount
                price = transaction.price

                self._cash_flow(
                    self._calculate_payout(
                        asset.price_multiplier,
                        amount,
                        old_price,
                        price,
                    ),
                )

                if amount + transaction.amount == 0:
                    del self._payout_last_sale_prices[asset]
                else:
                    self._payout_last_sale_prices[asset] = price
        else:
            self._cash_flow(-(transaction.price * transaction.amount))

        self.position_tracker.execute_transaction(transaction)

        # we only ever want the dict form from now on
        transaction_dict = transaction.to_dict()
        try:
            self._processed_transactions[transaction.dt].append(
                transaction_dict,
            )
        except KeyError:
            self._processed_transactions[transaction.dt] = [transaction_dict]

    def process_splits(self, splits):
        """Processes a list of splits by modifying any positions as needed.

        Parameters
        ----------
        splits: list[(Asset, float)]
            A list of splits. Each split is a tuple of (asset, ratio).
        """
        leftover_cash = self.position_tracker.handle_splits(splits)
        if leftover_cash > 0:
            self._cash_flow(leftover_cash)

    def process_order(self, order: Order):
        """Keep track of an order that was placed.

        Parameters
        ----------
        order : Order
            The order to record.
        """
        try:
            dt_orders = self._orders_by_modified[order.dt]
        except KeyError:
            self._orders_by_modified[order.dt] = OrderedDict(
                [
                    (order.id, order),
                ]
            )
            self._orders_by_id[order.id] = order
        else:
            self._orders_by_id[order.id] = dt_orders[order.id] = order
            # to preserve the order of the orders by modified date
            dt_orders.move_to_end(order.id, last=True)

        self._orders_by_id.move_to_end(order.id, last=True)

    def process_commission(self, commission: CommissionModel):
        """Process the commission.

        Parameters
        ----------
        commission : CommissionModel
            The commission being paid.
        """
        asset = commission["asset"]
        cost = commission["cost"]

        self.position_tracker.handle_commission(asset, cost)
        self._cash_flow(-cost)

    def close_position(self, asset: Asset, dt: datetime.datetime):
        txn = self.position_tracker.maybe_create_close_position_transaction(
            asset=asset,
            dt=dt,
        )
        if txn is not None:
            self.process_transaction(transaction=txn)

    def process_dividends(self, next_session, adjustment_reader):
        """Process dividends for the next session.

        This will earn us any dividends whose ex-date is the next session as
        well as paying out any dividends whose pay-date is the next session
        """
        position_tracker = self.position_tracker

        # Earn dividends whose ex_date is the next trading day. We need to
        # check if we own any of these stocks so we know to pay them out when
        # the pay date comes.
        held_sids = set(position_tracker.positions)
        if held_sids:
            cash_dividends = adjustment_reader.get_dividends_with_ex_date(
                held_sids, next_session, None  # self.data_bundle.asset_repository
            )
            stock_dividends = adjustment_reader.get_stock_dividends_with_ex_date(
                held_sids, next_session, None  # self.data_bundle.asset_repository
            )

            # Earning a dividend just marks that we need to get paid out on
            # the dividend's pay-date. This does not affect our cash yet.
            position_tracker.earn_dividends(
                cash_dividends,
                stock_dividends,
            )

        # Pay out the dividends whose pay-date is the next session. This does
        # affect out cash.
        self._cash_flow(
            position_tracker.pay_dividends(
                next_session,
            ),
        )

    def capital_change(self, change_amount: float):
        self.update_portfolio()
        # we update the cash and total value so this is not dirty
        self._portfolio.portfolio_value += change_amount
        self._portfolio.cash += change_amount

    def transactions(self, dt=None):
        """Retrieve the dict-form of all of the transactions in a given bar or
        for the whole simulation.

        Parameters
        ----------
        dt : datetime.datetime or None, optional
            The particular datetime to look up transactions for. If not passed,
            or None is explicitly passed, all of the transactions will be
            returned.

        Returns
        -------
        transactions : list[dict]
            The transaction information.
        """
        if dt is None:
            # flatten the by-day transactions
            return [
                txn
                for by_day in self._processed_transactions.values()
                for txn in by_day
            ]

        return self._processed_transactions.get(dt, [])

    def orders(self, dt=None):
        """Retrieve the dict-form of all of the orders in a given bar or for
        the whole simulation.

        Parameters
        ----------
        dt : datetime.datetime or None, optional
            The particular datetime to look up order for. If not passed, or
            None is explicitly passed, all of the orders will be returned.

        Returns
        -------
        orders : list[dict]
            The order information.
        """
        if dt is None:
            # orders by id is already flattened
            return [o.to_dict() for o in self._orders_by_id.values()]

        return [o.to_dict() for o in self._orders_by_modified.get(dt, {}).values()]

    @property
    def positions(self):
        return self.position_tracker.get_position_list()

    def _get_payout_total(self, positions):
        calculate_payout = self._calculate_payout
        payout_last_sale_prices = self._payout_last_sale_prices

        total = 0
        for asset, old_price in payout_last_sale_prices.items():
            position = positions[asset]
            payout_last_sale_prices[asset] = price = position.last_sale_price
            amount = position.amount
            total += calculate_payout(
                asset.price_multiplier,
                amount,
                old_price,
                price,
            )

        return total

    def update_portfolio(self) -> None:
        """Force a computation of the current portfolio state."""
        if not self._dirty_portfolio:
            return

        pt = self.position_tracker

        self._portfolio.positions = pt.get_positions()
        position_stats = pt.stats

        self._portfolio.positions_value = position_value = position_stats.net_value
        self._portfolio.positions_exposure = position_stats.net_exposure
        self._cash_flow(self._get_payout_total(pt.positions))

        start_value = self._portfolio.portfolio_value

        # update the new starting value
        self._portfolio.portfolio_value = end_value = self._portfolio.cash + position_value

        pnl = end_value - start_value
        if start_value != 0:
            returns = pnl / start_value
        else:
            returns = 0.00

        self._portfolio.pnl += pnl
        self._portfolio.returns = (1 + self._portfolio.returns) * (1 + returns) - 1

        # the portfolio has been fully synced
        self._dirty_portfolio = False

    @property
    def portfolio(self):
        """Compute the current portfolio.

        Notes
        -----
        This is cached, repeated access will not recompute the portfolio until
        the portfolio may have changed.
        """
        self.update_portfolio()
        return self._portfolio

    def calculate_period_stats(self):
        position_stats = self.position_tracker.stats
        portfolio_value = self.portfolio.portfolio_value

        if portfolio_value == 0:
            gross_leverage = net_leverage = np.inf
        else:
            gross_leverage = position_stats.gross_exposure / portfolio_value
            net_leverage = position_stats.net_exposure / portfolio_value

        return portfolio_value, gross_leverage, net_leverage

    @property
    def account(self):
        if self._dirty_account:
            portfolio = self.portfolio

            account = self._account

            # If no attribute is found in the ``_account_overrides`` resort to
            # the following default values. If an attribute is found use the
            # existing value. For instance, a exchange may provide updates to
            # these attributes. In this case we do not want to over write the
            # exchange values with the default values.
            account.settled_cash = portfolio.cash
            account.accrued_interest = 0.00
            account.buying_power = np.inf
            account.equity_with_loan = portfolio.portfolio_value
            account.total_positions_value = portfolio.portfolio_value - portfolio.cash
            account.total_positions_exposure = portfolio.positions_exposure
            account.regt_equity = portfolio.cash
            account.regt_margin = np.inf
            account.initial_margin_requirement = 0.00
            account.maintenance_margin_requirement = 0.00
            account.available_funds = portfolio.cash
            account.excess_liquidity = portfolio.cash
            account.cushion = (
                (portfolio.cash / portfolio.portfolio_value)
                if portfolio.portfolio_value
                else np.nan
            )
            account.day_trades_remaining = np.inf
            (
                account.net_liquidation,
                account.gross_leverage,
                account.net_leverage,
            ) = self.calculate_period_stats()

            account.leverage = account.gross_leverage

            # apply the overrides
            for k, v in self._account_overrides.items():
                setattr(account, k, v)

            # the account has been fully synced
            self._dirty_account = False

        return self._account



import enum


class SimulationEvent(enum.Enum):
    BAR = "BAR"
    SESSION_START = "SESSION_START"
    SESSION_END = "SESSION_END"
    EMISSION_RATE_END = "EMISSION_RATE_END"
    BEFORE_TRADING_START_BAR = "BEFORE_TRADING_START_BAR"



"""
Utilities for working with pandas objects.
"""

from contextlib import contextmanager
from copy import deepcopy
import warnings

import numpy as np
import pandas as pd


def explode(df):
    """Take a DataFrame and return a triple of

    (df.index, df.columns, df.values)
    """
    return df.index, df.columns, df.values


def nearest_unequal_elements(dts, dt):
    """Find values in ``dts`` closest but not equal to ``dt``.

    Returns a pair of (last_before, first_after).

    When ``dt`` is less than any element in ``dts``, ``last_before`` is None.
    When ``dt`` is greater any element in ``dts``, ``first_after`` is None.

    ``dts`` must be unique and sorted in increasing order.

    Parameters
    ----------
    dts : pd.DatetimeIndex
        Dates in which to search.
    dt : pd.Timestamp
        Date for which to find bounds.
    """
    if not dts.is_unique:
        raise ValueError("dts must be unique")

    if not dts.is_monotonic_increasing:
        raise ValueError("dts must be sorted in increasing order")

    if not len(dts):
        return None, None

    sortpos = dts.searchsorted(dt, side="left")
    try:
        sortval = dts[sortpos]
    except IndexError:
        # dt is greater than any value in the array.
        return dts[-1], None

    if dt < sortval:
        lower_ix = sortpos - 1
        upper_ix = sortpos
    elif dt == sortval:
        lower_ix = sortpos - 1
        upper_ix = sortpos + 1
    else:
        lower_ix = sortpos
        upper_ix = sortpos + 1

    lower_value = dts[lower_ix] if lower_ix >= 0 else None
    upper_value = dts[upper_ix] if upper_ix < len(dts) else None

    return lower_value, upper_value


def timedelta_to_integral_seconds(delta):
    """Convert a pd.Timedelta to a number of seconds as an int."""
    return int(delta.total_seconds())



@contextmanager
def ignore_pandas_nan_categorical_warning():
    with warnings.catch_warnings():
        # Pandas >= 0.18 doesn't like null-ish values in categories, but
        # avoiding that requires a broader change to how missing values are
        # handled in pipeline, so for now just silence the warning.
        warnings.filterwarnings(
            "ignore",
            category=FutureWarning,
        )
        yield


def categorical_df_concat(df_list, inplace=False):
    """Prepare list of pandas DataFrames to be used as input to pd.concat.
    Ensure any columns of type 'category' have the same categories across each
    dataframe.

    Parameters
    ----------
    df_list : list
        List of dataframes with same columns.
    inplace : bool
        True if input list can be modified. Default is False.

    Returns
    -------
    concatenated : df
        Dataframe of concatenated list.
    """

    if not inplace:
        df_list = deepcopy(df_list)

    # Assert each dataframe has the same columns/dtypes
    df = df_list[0]
    if not all([set(df.columns) == set(df_i.columns) for df_i in df_list[1:]]):
        raise ValueError("Input DataFrames must have the same columns.")

    categorical_columns = df.columns[df.dtypes == "category"]

    for col in categorical_columns:
        new_categories = _sort_set_none_first(
            _union_all(frame[col].cat.categories for frame in df_list)
        )

        with ignore_pandas_nan_categorical_warning():
            for df in df_list:
                df[col] = df[col].cat.set_categories(new_categories)

    return pd.concat(df_list)


def _union_all(iterables):
    """Union entries in ``iterables`` into a set."""
    return set().union(*iterables)


def _sort_set_none_first(set_):
    """Sort a set, sorting ``None`` before other elements, if present."""
    if None in set_:
        set_.remove(None)
        out = [None]
        out.extend(sorted(set_))
        set_.add(None)
        return out
    else:
        return sorted(set_)


def empty_dataframe(*columns):
    """Create an empty dataframe with columns of particular types.

    Parameters
    ----------
    *columns
        The (column_name, column_dtype) pairs.

    Returns
    -------
    typed_dataframe : pd.DataFrame
        The empty typed dataframe.

    Examples
    --------
    >>> df = empty_dataframe(
    ...     ('a', 'int64'),
    ...     ('b', 'float64'),
    ...     ('c', 'datetime64[ns]'),
    ... )

    >>> df
    Empty DataFrame
    Columns: [a, b, c]
    Index: []

    df.dtypes
    a             int64
    b           float64
    c    datetime64[ns]
    dtype: object
    """
    return pd.DataFrame(np.array([], dtype=list(columns)))



from functools import lru_cache

from sqlalchemy.orm import Mapped, relationship

from ziplime.assets.models.asset_model import AssetModel
from ziplime.assets.models.equity_symbol_mapping_model import EquitySymbolMappingModel


class EquityModel(AssetModel):
    __tablename__ = "equities"

    equity_symbol_mappings: Mapped[list[EquitySymbolMappingModel]] = relationship("EquitySymbolMappingModel")

    @lru_cache
    def get_symbol_by_exchange(self, exchange_name: str | None) -> str | None:
        if exchange_name is None:
            symbol_mapping = next(iter(self.equity_symbol_mappings), None)
        else:
            symbol_mapping = next(filter(lambda x: x.exchange == exchange_name, self.equity_symbol_mappings), None)
        if symbol_mapping:
            return symbol_mapping.symbol
        return None



import datetime
import math
from abc import abstractmethod

import structlog
from pandas import isnull

from ziplime.assets.entities.asset import Asset
from ziplime.assets.entities.equity import Equity
from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.errors import LiquidityExceeded
from ziplime.exchanges.exchange import Exchange
from ziplime.finance.shared import FinancialModelMeta
from ziplime.finance.domain.transaction import Transaction

SELL = 1 << 0
BUY = 1 << 1
STOP = 1 << 2
LIMIT = 1 << 3

SQRT_252 = math.sqrt(252)

DEFAULT_EQUITY_VOLUME_SLIPPAGE_BAR_LIMIT = 0.025
DEFAULT_FUTURE_VOLUME_SLIPPAGE_BAR_LIMIT = 0.05


class SlippageModel(metaclass=FinancialModelMeta):
    """Abstract base class for slippage models.

    Slippage models are responsible for the rates and prices at which orders
    fill during a simulation.

    To implement a new slippage model, create a subclass of
    :class:`~ziplime.finance.slippage.SlippageModel` and implement
    :meth:`process_order`.

    Methods
    -------
    process_order(data, order)

    Attributes
    ----------
    volume_for_bar : int
        Number of shares that have already been filled for the
        currently-filling asset in the current minute. This attribute is
        maintained automatically by the base class. It can be used by
        subclasses to keep track of the total amount filled if there are
        multiple open orders for a single asset.

    Notes
    -----
    Subclasses that define their own constructors should call
    ``super(<subclass name>, self).__init__()`` before performing other
    initialization.
    """

    # Asset types that are compatible with the given model.
    allowed_asset_types = (Equity, FuturesContract)

    def __init__(self):
        self._volume_for_bar = 0
        self._logger = structlog.get_logger(__name__)

    @property
    def volume_for_bar(self):
        return self._volume_for_bar

    @abstractmethod
    def process_order(self, exchange: Exchange, dt: datetime.datetime, order):
        """Compute the number of shares and price to fill for ``order`` in the
        current minute.

        Parameters
        ----------
        data : ziplime.protocol.BarData
            The data for the given bar.
        order : ziplime.finance.order.Order
            The order to simulate.

        Returns
        -------
        execution_price : float
            The price of the fill.
        execution_volume : int
            The number of shares that should be filled. Must be between ``0``
            and ``order.amount - order.filled``. If the amount filled is less
            than the amount remaining, ``order`` will remain open and will be
            passed again to this method in the next minute.

        Raises
        ------
        ziplime.finance.slippage.LiquidityExceeded
            May be raised if no more orders should be processed for the current
            asset during the current bar.

        Notes
        -----
        Before this method is called, :attr:`volume_for_bar` will be set to the
        number of shares that have already been filled for ``order.asset`` in
        the current minute.

        :meth:`process_order` is not called by the base class on bars for which
        there was no historical volume.
        """
        raise NotImplementedError("process_order")

    def simulate(self, exchange, assets: frozenset[Asset], orders_for_asset, current_dt: datetime.datetime):
        self._volume_for_bar = 0
        current_val = exchange.current(assets=assets, fields=frozenset({"close", "volume"}), dt=current_dt)

        volume_s = current_val["volume"]
        close_price_s = current_val["close"]
        if len(volume_s) == 0:
            self._logger.warning(f"No volume for {current_dt}, assets={[a.asset_name for a in assets]}")
            # volume is 0, since there is no volume our order couldn't have been executed
            return
        volume = volume_s[0]

        if volume == 0:
            return

        # can use the close price, since we verified there's volume in this
        # bar.
        price = close_price_s[0]

        # BEGIN
        #
        # Remove this block after fixing data to ensure volume always has
        # a corresponding price.
        if isnull(price):
            return
        # END

        for order in orders_for_asset:
            if order.open_amount == 0:
                continue

            order.check_triggers(price, current_dt)
            if not order.triggered:
                continue

            txn = None

            try:
                execution_price, execution_volume = self.process_order(exchange=exchange, dt=current_dt, order=order)

                if execution_price is not None:
                    # txn = create_transaction(
                    #     order,
                    #     data.current_dt,
                    #     execution_price,
                    #     execution_volume,
                    # )

                    txn = Transaction(
                        asset=order.asset,
                        amount=int(execution_volume),
                        dt=current_dt,
                        price=execution_price,
                        order_id=order.id,
                        exchange_name=exchange.name
                    )


            except LiquidityExceeded:
                break

            if txn:
                self._volume_for_bar += abs(txn.amount)
                yield order, txn

    def asdict(self):
        return self.__dict__



import datetime
import time
from typing import Any

import polars as pl
import structlog
from exchange_calendars import ExchangeCalendar, get_calendar

from ziplime.assets.services.asset_service import AssetService
from ziplime.constants.data_type import DataType
from ziplime.constants.period import Period
from ziplime.data.domain.data_bundle import DataBundle
from ziplime.data.services.data_bundle_source import DataBundleSource
from ziplime.data.services.bundle_registry import BundleRegistry
from ziplime.data.services.bundle_storage import BundleStorage
from ziplime.utils.class_utils import load_class
from ziplime.utils.date_utils import period_to_timedelta


class BundleService:

    def __init__(self, bundle_registry: BundleRegistry):
        self._bundle_registry = bundle_registry
        self._logger = structlog.get_logger(__name__)

    async def list_bundles(self) -> list[dict[str, Any]]:
        return await self._bundle_registry.list_bundles()

    async def ingest_custom_data_bundle(self, name: str,
                                        bundle_version: str,
                                        date_start: datetime.datetime,
                                        date_end: datetime.datetime,
                                        trading_calendar: ExchangeCalendar,
                                        symbols: list[str],
                                        data_bundle_source: DataBundleSource,
                                        frequency: datetime.timedelta | Period,
                                        data_frequency_use_window_end: bool,
                                        bundle_storage: BundleStorage,
                                        asset_service: AssetService,
                                        ):

        """Ingest data for a given bundle.        """
        self._logger.info(f"Ingesting custom bundle: name={name}, date_start={date_start}, date_end={date_end}, "
                          f"symbols={symbols}, frequency={frequency}")
        if date_start < trading_calendar.first_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError(
                f"Date start must be after first session of trading calendar. "
                f"First session is {trading_calendar.first_session.replace(tzinfo=trading_calendar.tz)} "
                f"and date start is {date_start}")

        if date_end > trading_calendar.last_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError(
                f"Date end must be before last session of trading calendar. "
                f"Last session is {trading_calendar.last_session.replace(tzinfo=trading_calendar.tz)} "
                f"and date end is {date_end}")

        data = await data_bundle_source.get_data(
            symbols=symbols,
            frequency=frequency,
            date_from=date_start,
            date_to=date_end
        )

        if data.is_empty():
            self._logger.warning(
                f"No data for symbols={symbols}, frequency={frequency}, date_start={date_start},"
                f"date_end={date_end} found. Skipping ingestion."
            )
            return

        # repair data
        all_bars = [
            s for s in pl.from_pandas(
                trading_calendar.sessions_minutes(start=date_start.replace(tzinfo=None),
                                                  end=date_end.replace(tzinfo=None)).tz_convert(trading_calendar.tz)
            ) if s >= date_start and s <= date_end
        ]

        required_sessions = pl.DataFrame({"date": all_bars}).group_by_dynamic(
            index_column="date", every=frequency
        ).agg()
        if data_frequency_use_window_end:
            if (
                    (type(frequency) is datetime.timedelta and frequency >= datetime.timedelta(days=1)) or
                    (type(frequency) is str and frequency in ["1d", "1w", "1mo", "1q", "1y"])
            ):
                last_row = required_sessions.tail(1).with_columns(
                    pl.col("date").dt.offset_by(frequency) - pl.duration(days=1))
                required_sessions = required_sessions.with_columns(
                    pl.col("date") - pl.duration(days=1)
                )[1:]

                required_sessions = pl.concat([required_sessions, last_row])
        required_columns = [
            "date"
        ]
        missing = [c for c in required_columns if c not in data.columns]

        if missing:
            raise ValueError(f"Ingested data is missing required columns: {missing}. Cannot ingest bundle.")
        if "symbol" not in data.columns and "sid" not in data.columns:
            raise ValueError(f"When ingesting custom bundle you must supply either a symbol or a sid column.")

        sid_id = "sid" in data.columns
        symbol_id = "symbol" in data.columns

        asset_identifiers = list(data["sid"].unique()) if sid_id else list(data["symbol"].unique())

        if sid_id:
            data = await self._backfill_symbol_data(data=data, asset_service=asset_service,
                                                    required_sessions=required_sessions)
        else:
            data = await self._backfill_sid_data(data=data, asset_service=asset_service,
                                                 required_sessions=required_sessions)

        data_bundle = DataBundle(name=name,
                                 start_date=date_start,
                                 end_date=date_end,
                                 trading_calendar=trading_calendar,
                                 frequency=frequency,
                                 original_frequency=frequency,
                                 data=data,
                                 timestamp=datetime.datetime.now(tz=trading_calendar.tz),
                                 version=bundle_version,
                                 data_type=DataType.CUSTOM
                                 )

        await self._bundle_registry.register_bundle(data_bundle=data_bundle, bundle_storage=bundle_storage)
        await bundle_storage.store_bundle(data_bundle=data_bundle)

        self._logger.info(f"Finished ingesting custom bundle_name={name}, bundle_version={bundle_version}")

    async def _backfill_sid_data(self, data: pl.DataFrame, asset_service: AssetService, required_sessions: pl.Series):

        unique_symbols = list(data["symbol"].unique())
        symbol_to_sid = {a.get_symbol_by_exchange(exchange_name=None): a.sid for a in
                         await asset_service.get_equities_by_symbols(unique_symbols)}
        data = data.with_columns(
            pl.lit(0).alias("sid"),
        )

        for symbol in unique_symbols:
            symbol_data = data.filter(symbol=symbol).with_columns(pl.col("date"))
            missing_sessions = sorted(set(required_sessions["date"]) - set(symbol_data["date"]))

            if len(missing_sessions) > 0:
                self._logger.warning(
                    f"Data for symbol {symbol} is missing on ticks ({len(missing_sessions)}): {[missing_session.isoformat() for missing_session in missing_sessions]}")
                new_rows_df = pl.DataFrame(
                    {"date": missing_sessions, "symbol": symbol},
                    schema_overrides={"date": data.schema["date"]}
                )
                # Concatenate with the original DataFrame
                data = pl.concat([data, new_rows_df], how="diagonal")
            missing_symbols = set(unique_symbols) - set(symbol_to_sid)
            if missing_symbols:
                raise ValueError(f"Symbols are missing in asset database: {missing_symbols}")

            data = data.with_columns(
                pl.col("symbol").replace(symbol_to_sid).cast(pl.Int64).alias("sid")
            ).sort(["sid", "date"])
        return data

    async def _backfill_symbol_data(self):
        pass

    async def ingest_market_data_bundle(self, name: str,
                                        bundle_version: str,
                                        date_start: datetime.datetime,
                                        date_end: datetime.datetime,
                                        trading_calendar: ExchangeCalendar,
                                        symbols: list[str],
                                        data_bundle_source: DataBundleSource,
                                        frequency: datetime.timedelta,
                                        bundle_storage: BundleStorage,
                                        asset_service: AssetService,
                                        forward_fill_missing_ohlcv_data: bool,
                                        ):

        """Ingest data for a given bundle.        """
        self._logger.info(f"Ingesting market data bundle: name={name}, date_start={date_start}, date_end={date_end}, "
                          f"symbols={symbols}, frequency={frequency}")
        start_duration = time.time()
        if date_start < trading_calendar.first_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError(
                f"Date start must be after first session of trading calendar. "
                f"First session is {trading_calendar.first_session.replace(tzinfo=trading_calendar.tz)} "
                f"and date start is {date_start}")

        if date_end > trading_calendar.last_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError(
                f"Date end must be before last session of trading calendar. "
                f"Last session is {trading_calendar.last_session.replace(tzinfo=trading_calendar.tz)} "
                f"and date end is {date_end}")

        data = await data_bundle_source.get_data(
            symbols=symbols,
            frequency=frequency,
            date_from=date_start,
            date_to=date_end
        )

        if data.is_empty():
            self._logger.warning(
                f"No data for symbols={symbols}, frequency={frequency}, date_from={date_start}, date_end={date_end} found. Skipping ingestion.")
            return

        required_columns = [
            "date", "symbol", "exchange", "exchange_country", "open", "high", "low", "close", "volume"
        ]
        missing = [c for c in required_columns if c not in data.columns]

        if missing:
            raise ValueError(f"Ingested data is missing required columns: {missing}. Cannot ingest bundle.")

        data = data.with_columns(
            pl.lit(0).alias("sid"),
            pl.lit(False).alias("backfilled")
        )
        # repair data
        all_bars = [
            s for s in pl.from_pandas(
                trading_calendar.sessions_minutes(start=date_start.replace(tzinfo=None),
                                                  end=date_end.replace(tzinfo=None)).tz_convert(trading_calendar.tz)
            ) if s >= date_start and s <= date_end
        ]
        required_sessions = pl.DataFrame({"date": all_bars, "close": 0.00}).group_by_dynamic(
            index_column="date", every=frequency
        ).agg()

        equities_by_exchange = data.select(
            "symbol", "exchange", "exchange_country"
        ).group_by("exchange", "exchange_country").agg(pl.col("symbol").unique())
        for row in equities_by_exchange.iter_rows(named=True):
            exchange_name = row["exchange"]
            exchange_country = row["exchange_country"]
            symbols = row["symbol"]
            equities = await asset_service.get_equities_by_symbols_and_exchange(symbols=symbols,
                                                                                exchange_name=exchange_name)
            symbol_to_sid = {e.get_symbol_by_exchange(exchange_name=exchange_name): e.sid for e in equities}

            for symbol in symbols:
                symbol_data = data.filter(symbol=symbol).with_columns(pl.col("date"))
                missing_sessions = sorted(set(required_sessions["date"]) - set(symbol_data["date"]))
                if len(missing_sessions) > 0:
                    self._logger.warning(
                        f"Data for symbol {symbol} is missing on ticks ({len(missing_sessions)}): {[missing_session.isoformat() for missing_session in missing_sessions]}")
                    new_rows_df = pl.DataFrame({"date": missing_sessions, "symbol": symbol, "exchange": exchange_name,
                                                "exchange_country": exchange_country},
                                               schema_overrides={"date": data.schema["date"]})

                    # Concatenate with the original DataFrame
                    data = pl.concat([data, new_rows_df], how="diagonal")
            missing_symbols = set(symbols) - set(symbol_to_sid)
            if missing_symbols:
                raise ValueError(f"Symbols are missing in asset database: {missing_symbols}")

            data = data.with_columns(
                pl.col("symbol").replace(symbol_to_sid).cast(pl.Int64).alias("sid")
            ).sort(["sid", "date"])
        if forward_fill_missing_ohlcv_data:
            data = data.with_columns(pl.col("close", "price").fill_null(strategy="forward"))
            data = data.with_columns(pl.col("high", "low", "open").fill_null(pl.col("price")))
            data = data.with_columns(pl.col("volume").fill_null(pl.lit(0.0)))

        data_bundle = DataBundle(name=name,
                                 start_date=date_start,
                                 end_date=date_end,
                                 trading_calendar=trading_calendar,
                                 frequency=frequency,
                                 original_frequency=frequency,
                                 data=data,
                                 timestamp=datetime.datetime.now(tz=trading_calendar.tz),
                                 version=bundle_version,
                                 data_type=DataType.MARKET_DATA
                                 )
        await self._bundle_registry.register_bundle(data_bundle=data_bundle, bundle_storage=bundle_storage)
        await bundle_storage.store_bundle(data_bundle=data_bundle)
        duration = time.time() - start_duration
        self._logger.info(f"Finished ingesting market data bundle_name={name}, bundle_version={bundle_version}."
                          f"Total duration: {duration:.2f} seconds", duration=duration)

    async def load_bundle(self, bundle_name: str, bundle_version: str | None,
                          symbols: list[str] | None = None,
                          start_date: datetime.datetime | None = None,
                          end_date: datetime.datetime | None = None,
                          frequency: datetime.timedelta | Period | None = None,
                          start_auction_delta: datetime.timedelta = None,
                          end_auction_delta: datetime.timedelta = None,
                          aggregations: list[pl.Expr] = None
                          ) -> DataBundle:
        self._logger.info(f"Loading bundle: bundle_name={bundle_name}, bundle_version={bundle_version}")

        bundle_metadata_start = time.time()

        bundle_metadata = await self._bundle_registry.load_bundle_metadata(bundle_name=bundle_name,
                                                                           bundle_version=bundle_version)
        if bundle_metadata is None:
            if bundle_version is None:
                raise ValueError(f"Bundle {bundle_name} not found.")
            else:
                raise ValueError(f"Bundle {bundle_name} with version {bundle_version} not found.")
        self._logger.info(f"Loaded bundle metadata in {time.time() - bundle_metadata_start} seconds")
        bundle_storage_class: BundleStorage = load_class(
            module_name='.'.join(bundle_metadata["bundle_storage_class"].split(".")[:-1]),
            class_name=bundle_metadata["bundle_storage_class"].split(".")[-1])

        bundle_storage = await bundle_storage_class.from_json(bundle_metadata["bundle_storage_data"])

        bundle_start_date = datetime.datetime.strptime(bundle_metadata["start_date"], "%Y-%m-%dT%H:%M:%SZ")
        trading_calendar = get_calendar(bundle_metadata["trading_calendar_name"],
                                        start=bundle_start_date - datetime.timedelta(days=30))
        bundle_start_date = bundle_start_date.replace(tzinfo=trading_calendar.tz)
        bundle_end_date = datetime.datetime.strptime(bundle_metadata["end_date"], "%Y-%m-%dT%H:%M:%SZ").replace(
            tzinfo=trading_calendar.tz)
        frequency_timedelta = datetime.timedelta(seconds=int(bundle_metadata["frequency_seconds"])) if bundle_metadata[
                                                                                                           "frequency_seconds"] is not None else None
        frequency_text = bundle_metadata.get("frequency_text", None)
        timestamp = datetime.datetime.strptime(bundle_metadata["timestamp"], "%Y-%m-%dT%H:%M:%SZ").replace(
            tzinfo=trading_calendar.tz)
        data_type = DataType(bundle_metadata["data_type"])
        bundle_frequency = frequency_timedelta or frequency_text

        if start_date is not None and start_date < bundle_start_date:
            raise ValueError(f"Start date {start_date} is before bundle start date {bundle_start_date}")
        if end_date is not None and end_date > bundle_end_date:
            raise ValueError(f"End date {end_date} is after bundle end date {bundle_end_date}")
        if frequency is not None and period_to_timedelta(frequency) < period_to_timedelta(bundle_frequency):
            raise ValueError(f"Requested frequency {frequency} is less than bundle frequency {bundle_frequency}")

        if start_auction_delta is not None and period_to_timedelta(start_auction_delta) < period_to_timedelta(
                bundle_frequency):
            raise ValueError(
                f"Requested start auction delta frequency {frequency} is less than bundle frequency {bundle_frequency}")

        if end_auction_delta is not None and period_to_timedelta(end_auction_delta) < period_to_timedelta(
                bundle_frequency):
            raise ValueError(
                f"Requested end auction delta frequency {frequency} is less than bundle frequency {bundle_frequency}")

        data_bundle = DataBundle(name=bundle_name,
                                 start_date=start_date or bundle_start_date,
                                 end_date=end_date or bundle_end_date,
                                 trading_calendar=trading_calendar,
                                 frequency=frequency or bundle_frequency,
                                 original_frequency=bundle_frequency,
                                 timestamp=timestamp,
                                 version=bundle_metadata["version"],
                                 data_type=data_type
                                 )
        bundle_data_load_start = time.time()

        data = await bundle_storage.load_data_bundle(data_bundle=data_bundle,
                                                     symbols=symbols,
                                                     start_date=start_date,
                                                     end_date=end_date,
                                                     frequency=frequency,
                                                     start_auction_delta=start_auction_delta,
                                                     end_auction_delta=end_auction_delta,
                                                     aggregations=aggregations
                                                     )
        load_duration = time.time() - bundle_data_load_start
        self._logger.info(f"Loaded data bundle in {load_duration:.2f} seconds",
                          duration=load_duration)
        data_bundle.data = data
        return data_bundle

    async def clean(self, bundle_name: str, before: datetime.datetime = None, after: datetime.datetime = None,
                    keep_last: bool = None):
        """Clean up data that was created with ``ingest`` or
        ``$ python -m ziplime ingest``

        Parameters
        ----------
        name : str
            The name of the bundle to remove data for.
        before : datetime, optional
            Remove data ingested before this date.
            This argument is mutually exclusive with: keep_last
        after : datetime, optional
            Remove data ingested after this date.
            This argument is mutually exclusive with: keep_last
        keep_last : int, optional
            Remove all but the last ``keep_last`` ingestions.
            This argument is mutually exclusive with:
              before
              after

        Returns
        -------
        cleaned : set[str]
            The names of the runs that were removed.

        Raises
        ------
        BadClean
            Raised when ``before`` and or ``after`` are passed with
            ``keep_last``. This is a subclass of ``ValueError``.
        """

        for bundle in await self._bundle_registry.list_bundles():
            self._delete_bundle(bundle)



"""
Helpers for downsampling code.
"""

from toolz import compose
from operator import attrgetter, methodcaller

from ziplime.utils.numpy_utils import changed_locations
from ziplime.utils.sharedoc import (
    templated_docstring,
    PIPELINE_DOWNSAMPLING_FREQUENCY_DOC,
)

_dt_to_period = {
    "year_start": attrgetter("year"),
    "quarter_start": attrgetter("quarter"),
    "month_start": attrgetter("month"),
    "week_start": compose(attrgetter("week"), methodcaller("isocalendar")),
}

SUPPORTED_DOWNSAMPLE_FREQUENCIES = frozenset(_dt_to_period)


@templated_docstring(frequency=PIPELINE_DOWNSAMPLING_FREQUENCY_DOC)
def select_sampling_indices(dates, frequency):
    """
    Choose entries from ``dates`` to use for downsampling at ``frequency``.

    Parameters
    ----------
    dates : pd.DatetimeIndex
        Dates from which to select sample choices.
    {frequency}

    Returns
    -------
    indices : np.array[int64]
        An array condtaining indices of dates on which samples should be taken.

        The resulting index will always include 0 as a sample index, and it
        will include the first date of each subsequent year/quarter/month/week,
        as determined by ``frequency``.

    Notes
    -----
    This function assumes that ``dates`` does not have large gaps.

    In particular, it assumes that the maximum distance between any two entries
    in ``dates`` is never greater than a year, which we rely on because we use
    ``np.diff(dates.<frequency>)`` to find dates where the sampling
    period has changed.
    """
    if frequency not in SUPPORTED_DOWNSAMPLE_FREQUENCIES:
        raise ValueError(
            "Frequency {frequency!r} is not supported. "
            "Supported frequencies are {supported_frequencies}.".format(
                frequency=frequency,
                supported_frequencies=sorted(SUPPORTED_DOWNSAMPLE_FREQUENCIES),
            )
        )
    return changed_locations(_dt_to_period[frequency](dates), include_first=True)



import datetime
from typing import Any

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger


class ConstantCumulativeRiskMetric:
    """A metric which does not change, ever.

    Notes
    -----
    This exists to maintain the existing structure of the perf packets. We
    should kill this as soon as possible.
    """

    def __init__(self, field, value):
        self._field = field
        self._value = value

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        packet["cumulative_risk_metrics"][self._field] = self._value

    def end_of_session(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                       exchanges: dict[str, Exchange]):
        packet["cumulative_risk_metrics"][self._field] = self._value



from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import UuidPK
from ziplime.core.db.base_model import BaseModel


class OrderModel(BaseModel):
    id: Mapped[UuidPK]

    base_asset_symbol: Mapped[str]
    quote_asset_symbol: Mapped[str]




from abc import abstractmethod
from logging import Logger

import structlog

from ziplime.core.base_task import BaseTask

from ziplime.exchanges.exchange import Exchange
from ziplime.trading.base_trading_algorithm import BaseTradingAlgorithm
from ziplime.trading.entities.orders.order_request import OrderRequest


class TradingSignal(BaseTask):
    """Trading signal represents action that needs to be done on the market.
    It can produce one or more orders depending on the signal type
    """

    def __init__(self, algorithm: BaseTradingAlgorithm, exchange: Exchange,
                 logger: Logger = structlog.get_logger(__name__)):
        super().__init__(logger=logger)
        self.algorithm = algorithm
        self.exchange = exchange

    @abstractmethod
    def get_orders_to_execute(self) -> list[OrderRequest]: ...



from datetime import datetime

from ziplime.finance.controls.account_control import AccountControl


class MinLeverage(AccountControl):
    """AccountControl representing a limit on the minimum leverage allowed
    by the algorithm after a threshold period of time.

    Parameters
    ----------
    min_leverage : float
        The gross leverage in decimal form.
    deadline : datetime
        The date the min leverage must be achieved by.

    For example, min_leverage=2 limits an algorithm to trading at minimum
    double the account value by the deadline date.
    """

    def __init__(self, min_leverage, deadline):
        super(MinLeverage, self).__init__(min_leverage=min_leverage, deadline=deadline)
        self.min_leverage = min_leverage
        self.deadline = deadline

    def validate(self, _portfolio, account, algo_datetime, _algo_current_data):
        """Make validation checks if we are after the deadline.
        Fail if the leverage is less than the min leverage.
        """
        if (
                algo_datetime > self.deadline.tz_localize(algo_datetime.tzinfo)
                and account.leverage < self.min_leverage
        ):
            self.fail()



from enum import Enum


class PositionMode(Enum):
    """
    Enum representing position modes in derivative trading.
    """
    HEDGE = "HEDGE"              # Can hold both long and short positions simultaneously
    ONEWAY = "ONEWAY"            # Can only hold either long or short position at a time
    PORTFOLIO_MARGIN = "PORTFOLIO_MARGIN"  # Uses portfolio risk assessment for margin calculation
    ISOLATED = "ISOLATED"        # Each position's margin is isolated from others
    CROSS = "CROSS"              # Positions share margin across the account



"""
Dataset representing OHLCV data.
"""

from ziplime.utils.numpy_utils import float64_dtype, categorical_dtype

from ..domain import US_EQUITIES
from .dataset import Column, DataSet


class EquityPricing(DataSet):
    """
    :class:`~ziplime.pipeline.data.DataSet` containing daily trading prices and
    volumes.
    """

    open = Column(float64_dtype, currency_aware=True)
    high = Column(float64_dtype, currency_aware=True)
    low = Column(float64_dtype, currency_aware=True)
    close = Column(float64_dtype, currency_aware=True)
    volume = Column(float64_dtype)
    currency = Column(categorical_dtype)


# Backwards compat alias.
USEquityPricing = EquityPricing.specialize(US_EQUITIES)



import datetime
from typing import Any

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger


class Returns:
    """Tracks the daily and cumulative returns of the algorithm."""

    # def _end_of_period(field, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
    #                    data_bundle: BundleData):
    #     packet[field]["returns"] = ledger.todays_returns
    #     packet["cumulative_perf"]["returns"] = ledger.portfolio.returns
    #     packet["cumulative_risk_metrics"][
    #         "algorithm_period_return"
    #     ] = ledger.portfolio.returns

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        packet["minute_perf"]["returns"] = ledger.todays_returns
        packet["cumulative_perf"]["returns"] = ledger.portfolio.returns
        packet["cumulative_risk_metrics"][
            "algorithm_period_return"
        ] = ledger.portfolio.returns

    def end_of_session(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                       exchanges: dict[str, Exchange]):
        packet["daily_perf"]["returns"] = ledger.todays_returns
        packet["cumulative_perf"]["returns"] = ledger.portfolio.returns
        packet["cumulative_risk_metrics"][
            "algorithm_period_return"
        ] = ledger.portfolio.returns
    # end_of_bar = partial(_end_of_period, "minute_perf")
    # end_of_session = partial(_end_of_period, "daily_perf")



import datetime
import operator as op
from typing import Any

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger


class SimpleLedgerField:
    """Emit the current value of a ledger field every bar or every session.

    Parameters
    ----------
    ledger_field : str
        The ledger field to read.
    packet_field : str, optional
        The name of the field to populate in the packet. If not provided,
        ``ledger_field`` will be used.
    """

    def __init__(self, ledger_field, packet_field=None):
        self._get_ledger_field = op.attrgetter(ledger_field)
        if packet_field is None:
            self._packet_field = ledger_field.rsplit(".", 1)[-1]
        else:
            self._packet_field = packet_field

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        packet["minute_perf"][self._packet_field] = self._get_ledger_field(
            ledger,
        )

    def end_of_session(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                       exchanges: dict[str, Exchange]):
        packet["daily_perf"][self._packet_field] = self._get_ledger_field(
            ledger,
        )



from typing import Any

from sqlalchemy import orm, JSON
from sqlalchemy.orm import DeclarativeBase

from ziplime.utils.string_utils import camel_case_to_snake_case


class BaseModel(DeclarativeBase):
    __name__: str

    type_annotation_map = {
        dict[str, Any]: JSON
    }

    # Generate __tablename__ automatically
    @orm.declared_attr
    def __tablename__(cls) -> str:
        return camel_case_to_snake_case(cls.__name__).lower()



from .iface import PipelineHooks
from .no import NoHooks
from .delegate import DelegatingHooks
from .progress import ProgressHooks


__all__ = [
    "PipelineHooks",
    "NoHooks",
    "DelegatingHooks",
    "ProgressHooks",
]



from ziplime.finance.shared import AllowedAssetMarker

from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.finance.slippage.slippage_model import SlippageModel


class FutureSlippageModel(SlippageModel, metaclass=AllowedAssetMarker):
    """Base class for slippage models which only support futures."""

    allowed_asset_types = (FuturesContract,)



from ziplime.core.entities.enums.task_status import TaskStatus
from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.order import Order
from ziplime.finance.domain.order_status import OrderStatus
from ziplime.finance.execution import MarketOrder, LimitOrder
from ziplime.trading.base_trading_algorithm import BaseTradingAlgorithm
from ziplime.trading.entities.orders.market_order_request import MarketOrderRequest
from ziplime.trading.entities.orders.order_request import OrderRequest
from ziplime.trading.entities.trading_signals.trading_signal import TradingSignal
from ziplime.trading.enums.order_type import OrderType


class OrderExecuteTradingSignal(TradingSignal):

    # order request should be replaced with OrderExecuteTradingSignalConfig ->
    def __init__(self, algorithm: BaseTradingAlgorithm, order: OrderRequest, exchange: Exchange):
        super().__init__(algorithm=algorithm, exchange=exchange)
        self.order = order

    def get_orders_to_execute(self) -> list[OrderRequest]:
        return [self.order]

    async def on_refresh_interval(self):
        """
        This method is responsible for controlling the task based on the status of the executor.

        :return: None
        """
        if self.status != TaskStatus.RUNNING:
            return None


        # Define variables needed before the match statement
        exchange = self.exchange

        match self.order.order_type:
            case OrderType.MARKET:
                execution_style = MarketOrder()
            case OrderType.LIMIT:
                execution_style = LimitOrder(limit_price=self.order.limit_price)
            case _:  # Default case to handle other order types
                # For now, default to market order if type is not recognized
                raise ValueError(f"Unsupported order type: {self.order.order_type}")

        order = Order(
            dt=self.algorithm.simulation_dt,
            asset=self.order.trading_pair.base_asset,
            amount=self.order.quantity,
            id=self.order.order_id,
            commission=0.00,
            filled=0,
            execution_style=execution_style,
            status=OrderStatus.OPEN,
            exchange_name=exchange.name
        )

        submitted_order = await self.algorithm.exchanges[self.order.exchange_name].submit_order(order)
        self.stop()
        self.algorithm.new_order_submitted(order=submitted_order)
        # getting price
        # self.algorithm.exchanges["xx"].get_price("")



import abc
from numpy import vectorize
from functools import partial, reduce
import operator
import pandas as pd
from collections import namedtuple
from toolz import groupby

from enum import IntEnum

from ziplime.assets.entities.asset import Asset
from ziplime.utils.numpy_utils import vectorized_is_element

Restriction = namedtuple("Restriction", ["asset", "effective_date", "state"])

RESTRICTION_STATES = IntEnum(
    "RESTRICTION_STATES",
    [
        "ALLOWED",
        "FROZEN",
    ],
    start=0,
)


class Restrictions(metaclass=abc.ABCMeta):
    """Abstract restricted list interface, representing a set of assets that an
    algorithm is restricted from trading.
    """

    @abc.abstractmethod
    def is_restricted(self, assets, dt):
        """Is the asset restricted (RestrictionStates.FROZEN) on the given dt?

        Parameters
        ----------
        asset : Asset of iterable of Assets
            The asset(s) for which we are querying a restriction
        dt : pd.Timestamp
            The timestamp of the restriction query

        Returns
        -------
        is_restricted : bool or pd.Series[bool] indexed by asset
            Is the asset or assets restricted on this dt?

        """
        raise NotImplementedError("is_restricted")

    def __or__(self, other_restriction):
        """Base implementation for combining two restrictions."""
        # If the right side is a _UnionRestrictions, defers to the
        # _UnionRestrictions implementation of `|`, which intelligently
        # flattens restricted lists
        if isinstance(other_restriction, _UnionRestrictions):
            return other_restriction | self
        return _UnionRestrictions([self, other_restriction])


class _UnionRestrictions(Restrictions):
    """A union of a number of sub restrictions.

    Parameters
    ----------
    sub_restrictions : iterable of Restrictions (but not _UnionRestrictions)
        The Restrictions to be added together

    Notes
    -----
    - Consumers should not construct instances of this class directly, but
      instead use the `|` operator to combine restrictions
    """

    def __new__(cls, sub_restrictions):
        # Filter out NoRestrictions and deal with resulting cases involving
        # one or zero sub_restrictions
        sub_restrictions = [
            r for r in sub_restrictions if not isinstance(r, NoRestrictions)
        ]
        if len(sub_restrictions) == 0:
            return NoRestrictions()
        elif len(sub_restrictions) == 1:
            return sub_restrictions[0]

        new_instance = super(_UnionRestrictions, cls).__new__(cls)
        new_instance.sub_restrictions = sub_restrictions
        return new_instance

    def __or__(self, other_restriction):
        """Overrides the base implementation for combining two restrictions, of
        which the left side is a _UnionRestrictions.
        """
        # Flatten the underlying sub restrictions of _UnionRestrictions
        if isinstance(other_restriction, _UnionRestrictions):
            new_sub_restrictions = (
                self.sub_restrictions + other_restriction.sub_restrictions
            )
        else:
            new_sub_restrictions = self.sub_restrictions + [other_restriction]

        return _UnionRestrictions(new_sub_restrictions)

    def is_restricted(self, assets, dt):
        if isinstance(assets, Asset):
            return any(r.is_restricted(assets, dt) for r in self.sub_restrictions)

        return reduce(
            operator.or_,
            (r.is_restricted(assets, dt) for r in self.sub_restrictions),
        )


class NoRestrictions(Restrictions):
    """A no-op restrictions that contains no restrictions."""

    def is_restricted(self, assets, dt):
        if isinstance(assets, Asset):
            return False
        return pd.Series(index=pd.Index(assets), data=False)


class StaticRestrictions(Restrictions):
    """Static restrictions stored in memory that are constant regardless of dt
    for each asset.

    Parameters
    ----------
    restricted_list : iterable of assets
        The assets to be restricted
    """

    def __init__(self, restricted_list):
        self._restricted_set = frozenset(restricted_list)

    def is_restricted(self, assets, dt):
        """An asset is restricted for all dts if it is in the static list."""
        if isinstance(assets, Asset):
            return assets in self._restricted_set
        return pd.Series(
            index=pd.Index(assets),
            data=vectorized_is_element(assets, self._restricted_set),
        )


class HistoricalRestrictions(Restrictions):
    """Historical restrictions stored in memory with effective dates for each
    asset.

    Parameters
    ----------
    restrictions : iterable of namedtuple Restriction
        The restrictions, each defined by an asset, effective date and state
    """

    def __init__(self, restrictions):
        # A dict mapping each asset to its restrictions, which are sorted by
        # ascending order of effective_date
        self._restrictions_by_asset = {
            asset: sorted(restrictions_for_asset, key=lambda x: x.effective_date)
            for asset, restrictions_for_asset in groupby(
                lambda x: x.asset, restrictions
            ).items()
        }

    def is_restricted(self, assets, dt):
        """Returns whether or not an asset or iterable of assets is restricted
        on a dt.
        """
        if isinstance(assets, Asset):
            return self._is_restricted_for_asset(assets, dt)

        is_restricted = partial(self._is_restricted_for_asset, dt=dt)
        return pd.Series(
            index=pd.Index(assets),
            data=vectorize(is_restricted, otypes=[bool])(assets),
        )

    def _is_restricted_for_asset(self, asset, dt):
        state = RESTRICTION_STATES.ALLOWED
        for r in self._restrictions_by_asset.get(asset, ()):
            r_effective_date = r.effective_date
            if r_effective_date.tzinfo is None:
                r_effective_date = r_effective_date.tz_localize(dt.tzinfo)
            if r_effective_date > dt:
                break
            state = r.state
        return state == RESTRICTION_STATES.FROZEN


# class SecurityListRestrictions(Restrictions):
#     """Restrictions based on a security list.
#
#     Parameters
#     ----------
#     restrictions : ziplime.utils.security_list.SecurityList
#         The restrictions defined by a SecurityList
#     """
#
#     def __init__(self, security_list_by_dt):
#         self.current_securities = security_list_by_dt.current_securities
#
#     def is_restricted(self, assets, dt):
#         securities_in_list = self.current_securities(dt)
#         if isinstance(assets, Asset):
#             return assets in securities_in_list
#         return pd.Series(
#             index=pd.Index(assets),
#             data=vectorized_is_element(assets, securities_in_list),
#         )



"""Canonical definitions of country code constants.
"""

from iso3166 import countries_by_name


def code(name):
    return countries_by_name[name].alpha2


class CountryCode:
    """A simple namespace of iso3166 alpha2 country codes."""

    ARGENTINA = code("ARGENTINA")
    AUSTRALIA = code("AUSTRALIA")
    AUSTRIA = code("AUSTRIA")
    BELGIUM = code("BELGIUM")
    BRAZIL = code("BRAZIL")
    CANADA = code("CANADA")
    CHILE = code("CHILE")
    CHINA = code("CHINA")
    COLOMBIA = code("COLOMBIA")
    CZECH_REPUBLIC = code("CZECHIA")
    DENMARK = code("DENMARK")
    FINLAND = code("FINLAND")
    FRANCE = code("FRANCE")
    GERMANY = code("GERMANY")
    GREECE = code("GREECE")
    HONG_KONG = code("HONG KONG")
    HUNGARY = code("HUNGARY")
    INDIA = code("INDIA")
    INDONESIA = code("INDONESIA")
    IRELAND = code("IRELAND")
    ISRAEL = code("ISRAEL")
    ITALY = code("ITALY")
    JAPAN = code("JAPAN")
    MALAYSIA = code("MALAYSIA")
    MEXICO = code("MEXICO")
    NETHERLANDS = code("NETHERLANDS")
    NEW_ZEALAND = code("NEW ZEALAND")
    NORWAY = code("NORWAY")
    PAKISTAN = code("PAKISTAN")
    PERU = code("PERU")
    PHILIPPINES = code("PHILIPPINES")
    POLAND = code("POLAND")
    PORTUGAL = code("PORTUGAL")
    RUSSIA = code("RUSSIAN FEDERATION")
    SINGAPORE = code("SINGAPORE")
    SOUTH_AFRICA = code("SOUTH AFRICA")
    SOUTH_KOREA = code("KOREA, REPUBLIC OF")
    SPAIN = code("SPAIN")
    SWEDEN = code("SWEDEN")
    SWITZERLAND = code("SWITZERLAND")
    TAIWAN = code("TAIWAN, PROVINCE OF CHINA")
    THAILAND = code("THAILAND")
    TURKEY = code("TÜRKIYE")
    UNITED_KINGDOM = code("UNITED KINGDOM OF GREAT BRITAIN AND NORTHERN IRELAND")
    UNITED_STATES = code("UNITED STATES OF AMERICA")



"""Caching utilities for ziplime"""



class Expired(Exception):
    """Marks that a :class:`CachedObject` has expired."""


ExpiredCachedObject = "ExpiredCachedObject"
AlwaysExpired = "AlwaysExpired"


class CachedObject:
    """A simple struct for maintaining a cached object with an expiration date.

    Parameters
    ----------
    value : object
        The object to cache.
    expires : datetime-like
        Expiration date of `value`. The cache is considered invalid for dates
        **strictly greater** than `expires`.

    Examples
    --------
    # >>> from pandas import Timestamp, Timedelta
    # >>> expires = Timestamp('2014', tz='UTC')
    # >>> obj = CachedObject(1, expires)
    # >>> obj.unwrap(expires - Timedelta('1 minute'))
    # 1
    # >>> obj.unwrap(expires)
    # 1
    # >>> obj.unwrap(expires + Timedelta('1 minute'))
    # ... # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    Expired: 2014-01-01 00:00:00+00:00
    """

    def __init__(self, value, expires):
        self._value = value
        self._expires = expires

    @classmethod
    def expired(cls):
        """Construct a CachedObject that's expired at any time."""
        return cls(ExpiredCachedObject, expires=AlwaysExpired)

    def unwrap(self, dt):
        """
        Get the cached value.

        Returns
        -------
        value : object
            The cached value.

        Raises
        ------
        Expired
            Raised when `dt` is greater than self.expires.
        """
        expires = self._expires
        if expires is AlwaysExpired or expires < dt:
            raise Expired(self._expires)
        return self._value

    def _unsafe_get_value(self):
        """You almost certainly shouldn't use this."""
        return self._value


class ExpiringCache:
    """A cache of multiple CachedObjects, which returns the wrapped the value
    or raises and deletes the CachedObject if the value has expired.

    Parameters
    ----------
    cache : dict-like, optional
        An instance of a dict-like object which needs to support at least:
        `__del__`, `__getitem__`, `__setitem__`
        If `None`, than a dict is used as a default.

    cleanup : callable, optional
        A method that takes a single argument, a cached object, and is called
        upon expiry of the cached object, prior to deleting the object. If not
        provided, defaults to a no-op.

    Examples
    --------
    >>> from pandas import Timestamp, Timedelta
    >>> expires = Timestamp('2014', tz='UTC')
    >>> value = 1
    >>> cache = ExpiringCache()
    >>> cache.set('foo', value, expires)
    >>> cache.get('foo', expires - Timedelta('1 minute'))
    1
    >>> cache.get('foo', expires + Timedelta('1 minute'))
    Traceback (most recent call last):
        ...
    KeyError: 'foo'
    """

    def __init__(self, cache=None, cleanup=lambda value_to_clean: None):
        if cache is not None:
            self._cache = cache
        else:
            self._cache = {}

        self.cleanup = cleanup

    def get(self, key, dt):
        """Get the value of a cached object.

        Parameters
        ----------
        key : any
            The key to lookup.
        dt : datetime
            The time of the lookup.

        Returns
        -------
        result : any
            The value for ``key``.

        Raises
        ------
        KeyError
            Raised if the key is not in the cache or the value for the key
            has expired.
        """
        try:
            return self._cache[key].unwrap(dt)
        except Expired as exc:
            self.cleanup(self._cache[key]._unsafe_get_value())
            del self._cache[key]
            raise KeyError(key) from exc

    def set(self, key, value, expiration_dt):
        """Adds a new key value pair to the cache.

        Parameters
        ----------
        key : any
            The key to use for the pair.
        value : any
            The value to store under the name ``key``.
        expiration_dt : datetime
            When should this mapping expire? The cache is considered invalid
            for dates **strictly greater** than ``expiration_dt``.
        """
        self._cache[key] = CachedObject(value, expiration_dt)



import abc

from ziplime.errors import AccountControlViolation


class AccountControl(metaclass=abc.ABCMeta):
    """Abstract base class representing a fail-safe control on the behavior of any
    algorithm.
    """

    def __init__(self, **kwargs):
        """Track any arguments that should be printed in the error message
        generated by self.fail.
        """
        self.__fail_args = kwargs

    @abc.abstractmethod
    def validate(self, _portfolio, _account, _algo_datetime, _algo_current_data):
        """On each call to handle data by TradingAlgorithm, this method should be
        called *exactly once* on each registered AccountControl object.

        If the check does not violate this AccountControl's restraint given
        the information in `portfolio` and `account`, this method should
        return None and have no externally-visible side-effects.

        If the desired order violates this AccountControl's constraint, this
        method should call self.fail().
        """
        raise NotImplementedError

    def fail(self):
        """Raise an AccountControlViolation with information about the failure."""
        raise AccountControlViolation(constraint=repr(self))

    def __repr__(self):
        return "{name}({attrs})".format(
            name=self.__class__.__name__, attrs=self.__fail_args
        )



import datetime
import time

import polars as pl
from exchange_calendars import ExchangeCalendar

from ziplime.trading.enums.simulation_event import SimulationEvent
from ziplime.gens.domain.trading_clock import TradingClock


class RealtimeClock(TradingClock):
    def __init__(self,
                 trading_calendar: ExchangeCalendar,
                 emission_rate: datetime.timedelta,
                 start_date: datetime.datetime | None = None,
                 end_date: datetime.datetime | None = None,
                 timedelta_diff_from_current_time: datetime.timedelta = None,
                 ):
        super().__init__(trading_calendar=trading_calendar, emission_rate=emission_rate)
        if start_date is not None and end_date is not None:
            if start_date >= end_date:
                raise ValueError("Period start falls after period end.")
            if start_date >= trading_calendar.last_session.replace(tzinfo=trading_calendar.tz):
                raise ValueError("Period start falls after the last known trading day.")
            if end_date <= trading_calendar.first_session.replace(tzinfo=trading_calendar.tz):
                raise ValueError("Period end falls before the first known trading day.")

            # chop off any minutes or hours on the given start and end dates,
            # as we only support session labels here (and we represent session
            # labels as midnight UTC).

        self.start_date = start_date
        self.end_date = end_date

        if self.start_date is None:
            self.start_date = datetime.datetime.now(tz=trading_calendar.tz)

        self.start_session = self.start_date.date()

        if self.end_date is None:
            self.end_date = datetime.datetime.now(tz=trading_calendar.tz) + datetime.timedelta(days=10)
        self.end_session = self.end_date.date()

        if trading_calendar.sessions_distance(self.start_session, self.end_session) < 1:
            raise Exception(
                f"There are no trading days between {self.start_session} and {self.end_session}")

        if not trading_calendar.is_session(self.start_session):
            # if the start date is not a valid session in this calendar,
            # push it forward to the first valid session
            self.start_session = trading_calendar.minute_to_session(
                self.start_session, direction="next"
            ).tz_localize(self.trading_calendar.tz).to_pydatetime().date()

        if not trading_calendar.is_session(self.end_session):
            # if the end date is not a valid session in this calendar,
            # pull it backward to the last valid session before the given
            # end date.
            self.end_session = trading_calendar.minute_to_session(
                self.end_session, direction="previous"
            ).tz_localize(self.trading_calendar.tz).to_pydatetime().date()

        self.first_open = trading_calendar.session_first_minute(
            self.start_session
        ).tz_convert(self.trading_calendar.tz).to_pydatetime()
        self.last_close = trading_calendar.session_close(
            self.end_session
        ).tz_convert(self.trading_calendar.tz).to_pydatetime()

        self.sessions = pl.Series(self.trading_calendar.sessions_in_range(
            self.start_session, self.end_session)
        ).dt.date()

        self.market_closes = pl.Series(
            self.trading_calendar.schedule.loc[self.sessions, "close"].dt.tz_convert(
                self.trading_calendar.tz))
        self.market_opens = pl.Series(
            self.trading_calendar.first_minutes.loc[self.sessions].dt.tz_convert(
                self.trading_calendar.tz))

        self.before_trading_start_minutes = self.market_opens - datetime.timedelta(minutes=46)

        if timedelta_diff_from_current_time is None:
            self.timedelta_diff_from_current_time = datetime.timedelta(seconds=0)
        else:
            self.timedelta_diff_from_current_time = timedelta_diff_from_current_time
        self.before_trading_start_minutes = self.before_trading_start_minutes
        self.before_trading_start_bar_yielded = len(self.before_trading_start_minutes) * [False]
        self.market_closes_yielded = len(self.market_closes) * [False]
        self.market_opens_yielded = len(self.market_opens) * [False]

    def _sleep_and_increase_time(self, sleep_seconds: int) -> datetime.datetime:
        time.sleep(sleep_seconds)
        now = datetime.datetime.now(tz=self.trading_calendar.tz) + self.timedelta_diff_from_current_time
        return now

    def __iter__(self):
        current_time = datetime.datetime.now(tz=self.trading_calendar.tz) + self.timedelta_diff_from_current_time
        last_bar_emit: datetime.datetime | None = None
        while current_time <= self.market_closes[-1]:
            current_session_index = self.sessions.index_of(current_time.date())
            if current_session_index is None:
                # happens if current time is not in trading hours, so we need to wait till market hours
                self._logger.info(f"Current time {current_time} is not in trading hours, waiting till market hours.")
                current_time = self._sleep_and_increase_time(sleep_seconds=1)
                continue
            if current_time >= self.before_trading_start_minutes[current_session_index] and not \
                    self.before_trading_start_bar_yielded[current_session_index]:
                self.before_trading_start_bar_yielded[current_session_index] = True
                yield current_time, SimulationEvent.BEFORE_TRADING_START_BAR
            elif current_time < self.market_opens[current_session_index]:
                current_time = self._sleep_and_increase_time(sleep_seconds=1)
            elif self.market_opens[current_session_index] <= current_time <= self.market_closes[
                current_session_index] and not self.market_opens_yielded[current_session_index]:
                self.market_opens_yielded[current_session_index] = True
                yield current_time.date(), SimulationEvent.SESSION_START
            elif self.market_opens[current_session_index] <= current_time <= self.market_closes[current_session_index]:
                if last_bar_emit is None or current_time - last_bar_emit >= self.emission_rate:
                    last_bar_emit = current_time
                    yield current_time, SimulationEvent.BAR
                    current_time = datetime.datetime.now(
                        tz=self.trading_calendar.tz) + self.timedelta_diff_from_current_time
                    yield current_time, SimulationEvent.EMISSION_RATE_END
                else:
                    current_time = self._sleep_and_increase_time(sleep_seconds=1)
            elif current_time > self.market_closes[current_session_index]:
                if self.market_closes_yielded[current_session_index]:
                    self._logger.info(
                        f"Current time {current_time} is after trading hours, waiting till market hours.")
                    current_time = self._sleep_and_increase_time(sleep_seconds=1)
                else:
                    last_bar_emit = current_time
                    self.market_closes_yielded[current_session_index] = True
                    yield current_time.date(), SimulationEvent.SESSION_END
            else:
                # We should never end up in this branch
                raise RuntimeError("Invalid state in RealtimeClock")



# Import all the models, so that Base has them before being
# imported by Alembic
from .base_model import BaseModel # noqa
from ziplime.assets.models.exchange_info import ExchangeInfo # noqa
from ziplime.assets.models.asset_router import AssetRouter # noqa
from ziplime.assets.models.currency_model import CurrencyModel # noqa
from ziplime.assets.models.currency_symbol_mapping_model import CurrencySymbolMappingModel # noqa
from ziplime.assets.models.commodity_model import CommodityModel # noqa
from ziplime.assets.models.dividend import Dividend # noqa
from ziplime.assets.models.divident_payout import DividendPayout # noqa
from ziplime.assets.models.equity_model import EquityModel # noqa
from ziplime.assets.models.equity_supplementary_mapping import EquitySupplementaryMapping # noqa
from ziplime.assets.models.equity_symbol_mapping_model import EquitySymbolMappingModel # noqa
from ziplime.assets.models.futures_contract_model import FuturesContractModel # noqa
from ziplime.assets.models.futures_root_symbol import FuturesRootSymbol # noqa
from ziplime.assets.models.merger import Merger # noqa
from ziplime.assets.models.split import Split # noqa
from ziplime.assets.models.stock_divident_payout import StockDividendPayout # noqa
from ziplime.assets.models.symbols_universe import SymbolsUniverseModel # noqa
from ziplime.assets.models.symbols_universe_asset import SymbolsUniverseAssetModel # noqa
# from ziplime.assets.models.trading_pair import TradingPair # noqa



import re
from collections import namedtuple

import logging
import pandas as pd
import sqlalchemy as sa
import structlog
from toolz import (
    concatv,
    curry,
    groupby,
    sliding_window,
    valmap,
)

import numpy as np

from .domain.continuous_future import ContinuousFuture

log = structlog.get_logger("assets.py")

# A set of fields that need to be converted to timestamps in UTC
_asset_timestamp_fields = frozenset(
    {
        "start_date",
        "end_date",
        "first_traded",
        "notice_date",
        "expiration_date",
        "auto_close_date",
    }
)
_delimited_symbol_default_triggers = frozenset({np.nan, None, ""})
_delimited_symbol_delimiters_regex = re.compile(r"[./\-_]")

OwnershipPeriod = namedtuple("OwnershipPeriod", "start end sid value")
SYMBOL_COLUMNS = frozenset(
    {
        "symbol",
        "company_symbol",
        "share_class_symbol",
    }
)


def split_delimited_symbol(symbol):
    """
    Takes in a symbol that may be delimited and splits it in to a company
    symbol and share class symbol. Also returns the fuzzy symbol, which is the
    symbol without any fuzzy characters at all.

    Parameters
    ----------
    symbol : str
        The possibly-delimited symbol to be split

    Returns
    -------
    company_symbol : str
        The company part of the symbol.
    share_class_symbol : str
        The share class part of a symbol.
    """
    # return blank strings for any bad fuzzy symbols, like NaN or None
    if symbol in _delimited_symbol_default_triggers:
        return "", ""

    symbol = symbol.upper()

    split_list = re.split(
        pattern=_delimited_symbol_delimiters_regex,
        string=symbol,
        maxsplit=1,
    )

    # Break the list up in to its two components, the company symbol and the
    # share class symbol
    company_symbol = split_list[0]
    if len(split_list) > 1:
        share_class_symbol = split_list[1]
    else:
        share_class_symbol = ""

    return company_symbol, share_class_symbol


def merge_ownership_periods(mappings):
    """Given a dict of mappings where the values are lists of
    OwnershipPeriod objects, returns a dict with the same structure with
    new OwnershipPeriod objects adjusted so that the periods have no
    gaps.

    Orders the periods chronologically, and pushes forward the end date
    of each period to match the start date of the following period. The
    end date of the last period pushed forward to the max Timestamp.
    """
    return valmap(
        lambda v: tuple(
            OwnershipPeriod(
                a.start,
                b.start,
                a.sid,
                a.value,
            )
            for a, b in sliding_window(
                2,
                concatv(
                    sorted(v),
                    # concat with a fake ownership object to make the last
                    # end date be max timestamp
                    [
                        OwnershipPeriod(
                            pd.Timestamp.max,
                            None,
                            None,
                            None,
                        )
                    ],
                ),
            )
        ),
        mappings,
    )


def _build_ownership_map_from_rows(rows, key_from_row, value_from_row):
    mappings = {}
    for row in rows:
        mappings.setdefault(
            key_from_row(row),
            [],
        ).append(
            OwnershipPeriod(
                # TODO FIX TZ MESS
                # pd.Timestamp(row.start_date, unit="ns", tz="utc"),
                # pd.Timestamp(row.end_date, unit="ns", tz="utc"),
                pd.Timestamp(row.start_date, unit="ns", tz='UTC').to_pydatetime(),
                pd.Timestamp(row.end_date, unit="ns", tz='UTC').to_pydatetime(),
                row.sid,
                value_from_row(row),
            ),
        )

    return merge_ownership_periods(mappings)


def build_ownership_map(conn, table, key_from_row, value_from_row):
    """Builds a dict mapping to lists of OwnershipPeriods, from a db table."""
    return _build_ownership_map_from_rows(
        conn.execute(sa.select(table.c)).fetchall(),
        key_from_row,
        value_from_row,
    )


def build_grouped_ownership_map(conn, table, key_from_row, value_from_row, group_key):
    """Builds a dict mapping group keys to maps of keys to lists of
    OwnershipPeriods, from a db table.
    """

    grouped_rows = groupby(
        group_key,
        conn.execute(sa.select(table.c)).fetchall(),
    )
    return {
        key: _build_ownership_map_from_rows(
            rows,
            key_from_row,
            value_from_row,
        )
        for key, rows in grouped_rows.items()
    }


@curry
def _filter_kwargs(names, dict_):
    """Filter out kwargs from a dictionary.

    Parameters
    ----------
    names : set[str]
        The names to select from ``dict_``.
    dict_ : dict[str, any]
        The dictionary to select from.

    Returns
    -------
    kwargs : dict[str, any]
        ``dict_`` where the keys intersect with ``names`` and the values are
        not None.
    """
    return {k: v for k, v in dict_.items() if k in names and v is not None}


asset_args = frozenset({
    'sid',
    'symbol',
    'asset_name',
    'start_date',
    'end_date',
    'first_traded',
    'auto_close_date',
    'tick_size',
    'multiplier',
    'exchange_info',
})
_filter_future_kwargs = _filter_kwargs(asset_args)
_filter_equity_kwargs = _filter_kwargs(asset_args)


def _convert_asset_timestamp_fields(dict_):
    """Takes in a dict of Asset init args and converts dates to pd.Timestamps"""
    for key in _asset_timestamp_fields & dict_.keys():
        # TODO FIX TZ MESS
        # value = pd.Timestamp(dict_[key], tz="UTC")
        value = pd.Timestamp(dict_[key], tz=None).to_pydatetime().date()
        dict_[key] = None if pd.isnull(value) else value
    return dict_


SID_TYPE_IDS = {
    # Asset would be 0,
    ContinuousFuture: 1,
}

CONTINUOUS_FUTURE_ROLL_STYLE_IDS = {
    "calendar": 0,
    "volume": 1,
}

CONTINUOUS_FUTURE_ADJUSTMENT_STYLE_IDS = {
    None: 0,
    "div": 1,
    "add": 2,
}


def _encode_continuous_future_sid(root_symbol, offset, roll_style, adjustment_style):
    # Generate a unique int identifier
    values = (
        SID_TYPE_IDS[ContinuousFuture],
        offset,
        *[ord(x) for x in root_symbol.upper()],
        CONTINUOUS_FUTURE_ROLL_STYLE_IDS[roll_style],
        CONTINUOUS_FUTURE_ADJUSTMENT_STYLE_IDS[adjustment_style],
    )
    return int("".join([str(x) for x in values]))


Lifetimes = namedtuple("Lifetimes", "sid start end")



from dataclasses import dataclass

@dataclass
class Account:
    """
    The account object tracks information about the trading account. The
    values are updated as the algorithm runs and its keys remain unchanged.
    If connected to a exchange, one can update these values with the trading
    account values as reported by the exchange.
    """

    settled_cash: float
    accrued_interest: float
    buying_power: float
    equity_with_loan: float
    total_positions_value: float
    total_positions_exposure: float
    regt_equity: float
    regt_margin: float
    initial_margin_requirement: float
    maintenance_margin_requirement: float
    available_funds: float
    excess_liquidity: float
    cushion: float
    day_trades_remaining: float
    leverage: float
    net_leverage: float
    net_liquidation: float



"""
Base class for Pipeline API data loaders.
"""


class PipelineLoader:
    """Interface for PipelineLoaders."""

    def load_adjusted_array(self, domain, columns, dates, sids, mask):
        """
        Load data for ``columns`` as AdjustedArrays.

        Parameters
        ----------
        domain : ziplime.pipeline.domain.Domain
            The domain of the pipeline for which the requested data must be
            loaded.
        columns : list[ziplime.pipeline.data.dataset.BoundColumn]
            Columns for which data is being requested.
        dates : pd.DatetimeIndex
            Dates for which data is being requested.
        sids : pd.Int64Index
            Asset identifiers for which data is being requested.
        mask : np.array[ndim=2, dtype=bool]
            Boolean array of shape (len(dates), len(sids)) indicating dates on
            which we believe the requested assets were alive/tradeable. This is
            used for optimization by some loaders.

        Returns
        -------
        arrays : dict[BoundColumn -> ziplime.lib.adjusted_array.AdjustedArray]
            Map from column to an AdjustedArray representing a point-in-time
            rolling view over the requested dates for the requested sids.
        """

    @property
    def currency_aware(self):
        """Whether or not this loader supports currency-conversions.

        By default, assume that loaders to not support currency conversions.
        """
        return False



import datetime
import polars as pl

from ziplime.assets.entities.asset import Asset
from ziplime.constants.data_type import DataType
from ziplime.constants.period import Period
from ziplime.utils.date_utils import period_to_timedelta


class DataSource:

    def __init__(self, name: str, start_date: datetime.date, end_date: datetime.date,
                 frequency: datetime.timedelta | Period,
                 original_frequency: datetime.timedelta | Period,
                 data_type: DataType,
                 aggregation_specification: dict[str, str] = None):
        self.name = name
        self.start_date = start_date
        self.end_date = end_date
        self.frequency = frequency
        self.frequency_td = period_to_timedelta(self.frequency)
        self.data_type = data_type
        self.aggregation_specification = aggregation_specification
        self.original_frequency = original_frequency

    def get_dataframe(self) -> pl.DataFrame:
        return self.data

    def get_data_by_date(self, fields: frozenset[str],
                         from_date: datetime.datetime,
                         to_date: datetime.datetime,
                         frequency: datetime.timedelta | Period,
                         assets: frozenset[Asset],
                         include_bounds: bool,
                         ) -> pl.DataFrame:

        cols = set(fields.union({"date", "sid"}))
        if include_bounds:
            df = self.get_dataframe().select(pl.col(col) for col in cols).filter(
                pl.col("date") <= to_date,
                pl.col("date") >= from_date,
                pl.col("sid").is_in([asset.sid for asset in assets])
            ).group_by(pl.col("sid")).all()
        else:
            df = self.get_dataframe().select(pl.col(col) for col in cols).filter(
                pl.col("date") < to_date,
                pl.col("date") > from_date,
                pl.col("sid").is_in([asset.sid for asset in assets])).group_by(pl.col("sid")).all()
        if self.frequency < frequency:
            df = df.group_by_dynamic(
                index_column="date", every=frequency, by="sid").agg(pl.col(field).last() for field in fields)
        return df.sort(by="date")

    # @lru_cache(maxsize=100)
    def get_data_by_limit(self, fields: frozenset[str] | None,
                          limit: int,
                          end_date: datetime.datetime,
                          frequency: datetime.timedelta | Period,
                          assets: frozenset[Asset],
                          include_end_date: bool,
                          ) -> pl.DataFrame:
        frequency_td = period_to_timedelta(frequency)
        total_bar_count = limit
        if end_date > self.end_date:
            return self.get_missing_data_by_limit(frequency=frequency, assets=assets, fields=fields,
                                                  limit=limit, include_end_date=include_end_date,
                                                  end_date=end_date
                                                  )  # pl.DataFrame() # we have missing data

        if self.frequency_td < frequency_td:
            multiplier = int(frequency_td / self.frequency_td)
            total_bar_count = limit * multiplier
        df = self.get_dataframe()
        if fields is None:
            fields = frozenset(df.columns)
        cols = list(fields.union({"date", "sid"}))

        if include_end_date:
            df_raw = self.get_dataframe().select(pl.col(col) for col in cols).filter(
                pl.col("date") <= end_date,
                pl.col("sid").is_in([asset.sid for asset in assets])
            ).group_by(pl.col("sid")).tail(total_bar_count).sort(by="date")
        else:
            df_raw = self.get_dataframe().select(pl.col(col) for col in cols).filter(
                pl.col("date") < end_date,
                pl.col("sid").is_in([asset.sid for asset in assets])).group_by(pl.col("sid")).tail(
                total_bar_count).sort(by="date")

        if self.frequency_td < frequency_td:
            df = df_raw.group_by_dynamic(
                index_column="date", every=frequency, by="sid").agg(pl.col(field).last() for field in fields).tail(
                limit)
            return df
        return df_raw

    def get_spot_value(self, assets: frozenset[Asset], fields: frozenset[str], dt: datetime.datetime,
                       frequency: datetime.timedelta):
        """Public API method that returns a scalar value representing the value
        of the desired asset's field at either the given dt.

        Parameters
        ----------
        assets : Asset, ContinuousFuture, or iterable of same.
            The asset or assets whose data is desired.
        field : {'open', 'high', 'low', 'close', 'volume',
                 'price', 'last_traded'}
            The desired field of the asset.
        dt : datetime.datetime
            The timestamp for the desired value.
        data_frequency : str
            The frequency of the data to query; i.e. whether the data is
            'daily' or 'minute' bars

        Returns
        -------
        value : float, int, or datetime.datetime
            The spot value of ``field`` for ``asset`` The return type is based
            on the ``field`` requested. If the field is one of 'open', 'high',
            'low', 'close', or 'price', the value will be a float. If the
            ``field`` is 'volume' the value will be a int. If the ``field`` is
            'last_traded' the value will be a Timestamp.
        """
        # print(f"get spot value: {assets}, {fields}, {dt}")
        df_raw = self.get_data_by_limit(
            fields=fields,
            limit=1,
            end_date=dt,
            frequency=frequency,
            assets=assets,
            include_end_date=True,
        )
        return df_raw



#
# Copyright 2014 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import datetime
import uuid
from sys import float_info
from numpy import isfinite
import ziplime.utils.math_utils as zp_math
from ziplime.assets.entities.asset import Asset
from ziplime.errors import BadOrderParameters
from ziplime.trading.entities.orders.market_order_request import MarketOrderRequest
from ziplime.trading.entities.trading_pair import TradingPair
from ziplime.trading.enums.order_side import OrderSide
from ziplime.trading.enums.order_type import OrderType
from ziplime.utils.compat import consistent_round


class ExecutionStyle(metaclass=abc.ABCMeta):
    """Base class for order execution styles."""

    _exchange = None

    @abc.abstractmethod
    def get_limit_price(self, is_buy):
        """
        Get the limit price for this order.
        Returns either None or a numerical value >= 0.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_stop_price(self, is_buy):
        """
        Get the stop price for this order.
        Returns either None or a numerical value >= 0.
        """
        raise NotImplementedError

    @property
    def exchange(self):
        """
        The exchange to which this order should be routed.
        """
        return self._exchange

    def to_order_type(self) -> OrderType:
        raise NotImplementedError("to_order_type not implemented for ExecutionStyle")

    async def to_order_request(self, base_asset: Asset, quote_asset: Asset,
                               quantity: int,
                               exchange_name: str,
                               creation_dt: datetime.datetime,
                               ): ...


class MarketOrder(ExecutionStyle):
    """
    Execution style for orders to be filled at current market price.

    This is the default for orders placed with :func:`~ziplime.api.order`.
    """

    def __init__(self, exchange=None):
        self._exchange = exchange

    def get_limit_price(self, is_buy: bool):
        return None

    def get_stop_price(self, is_buy: bool):
        return None

    def to_order_type(self) -> OrderType:
        return OrderType.MARKET

    async def to_order_request(self, base_asset: Asset, quote_asset: Asset,
                               quantity: int,
                               exchange_name: str,
                               creation_dt: datetime.datetime,
                               ) -> MarketOrderRequest:
        order_req = MarketOrderRequest(
            order_id=uuid.uuid4().hex,
            trading_pair=TradingPair(base_asset=base_asset,
                                     quote_asset=quote_asset,
                                     exchange_name=exchange_name),
            order_side=OrderSide.BUY if quantity > 0 else OrderSide.SELL,
            quantity=float(quantity),
            exchange_name=exchange_name,
            creation_date=creation_dt
        )
        return order_req

    def __str__(self):
        return "MarketOrder()"


class LimitOrder(ExecutionStyle):
    """
    Execution style for orders to be filled at a price equal to or better than
    a specified limit price.

    Parameters
    ----------
    limit_price : float
        Maximum price for buys, or minimum price for sells, at which the order
        should be filled.
    """

    def __init__(self, limit_price, asset=None, exchange=None):
        check_stoplimit_prices(price=limit_price, label="limit")

        self.limit_price = limit_price
        self._exchange = exchange
        self.asset = asset

    def get_limit_price(self, is_buy):
        return asymmetric_round_price(
            self.limit_price,
            is_buy,
            tick_size=(0.01 if self.asset is None else self.asset.tick_size),
        )

    def get_stop_price(self, is_buy):
        return None

    def to_order_type(self) -> OrderType:
        return OrderType.LIMIT

    def __str__(self):
        return f"LimitOrder(limit_price={self.limit_price})"


class StopOrder(ExecutionStyle):
    """
    Execution style representing a market order to be placed if market price
    reaches a threshold.

    Parameters
    ----------
    stop_price : float
        Price threshold at which the order should be placed. For sells, the
        order will be placed if market price falls below this value. For buys,
        the order will be placed if market price rises above this value.
    """

    def __init__(self, stop_price, asset=None, exchange=None):
        check_stoplimit_prices(price=stop_price, label="stop")

        self.stop_price = stop_price
        self._exchange = exchange
        self.asset = asset

    def get_limit_price(self, _is_buy):
        return None

    def get_stop_price(self, is_buy):
        return asymmetric_round_price(
            self.stop_price,
            not is_buy,
            tick_size=(0.01 if self.asset is None else self.asset.tick_size),
        )

    def to_order_type(self) -> OrderType:
        return OrderType.STOP

    def __str__(self):
        return f"StopOrder(stop_price={self.stop_price})"


class StopLimitOrder(ExecutionStyle):
    """
    Execution style representing a limit order to be placed if market price
    reaches a threshold.

    Parameters
    ----------
    limit_price : float
        Maximum price for buys, or minimum price for sells, at which the order
        should be filled, if placed.
    stop_price : float
        Price threshold at which the order should be placed. For sells, the
        order will be placed if market price falls below this value. For buys,
        the order will be placed if market price rises above this value.
    """

    def __init__(self, limit_price, stop_price, asset=None, exchange=None):
        check_stoplimit_prices(price=limit_price, label="limit")
        check_stoplimit_prices(price=stop_price, label="stop")

        self.limit_price = limit_price
        self.stop_price = stop_price
        self._exchange = exchange
        self.asset = asset

    def get_limit_price(self, is_buy):
        return asymmetric_round_price(
            self.limit_price,
            is_buy,
            tick_size=(0.01 if self.asset is None else self.asset.tick_size),
        )

    def get_stop_price(self, is_buy):
        return asymmetric_round_price(
            self.stop_price,
            not is_buy,
            tick_size=(0.01 if self.asset is None else self.asset.tick_size),
        )

    def to_order_type(self) -> OrderType:
        return OrderType.STOP_LIMIT

    def __str__(self):
        return f"StopLimitOrder(limit_price={self.limit_price}, stop_price={self.stop_price})"


def asymmetric_round_price(price: float, prefer_round_down: bool, tick_size: float, diff: float = 0.95):
    """
    Asymmetric rounding function for adjusting prices to the specified number
    of places in a way that "improves" the price. For limit prices, this means
    preferring to round down on buys and preferring to round up on sells.
    For stop prices, it means the reverse.

    If prefer_round_down == True:
        When .05 below to .95 above a specified decimal place, use it.
    If prefer_round_down == False:
        When .95 below to .05 above a specified decimal place, use it.

    In math-speak:
    If prefer_round_down: [<X-1>.0095, X.0195) -> round to X.01.
    If not prefer_round_down: (<X-1>.0005, X.0105] -> round to X.01.
    """
    precision = zp_math.number_of_decimal_places(tick_size)
    multiplier = int(tick_size * (10 ** precision))
    diff -= 0.5  # shift the difference down
    diff *= 10.0 ** -precision  # adjust diff to precision of tick size
    diff *= multiplier  # adjust diff to value of tick_size

    # Subtracting an epsilon from diff to enforce the open-ness of the upper
    # bound on buys and the lower bound on sells.  Using the actual system
    # epsilon doesn't quite get there, so use a slightly less epsilon-ey value.
    epsilon = float_info.epsilon * 10
    diff = diff - epsilon

    # relies on rounding half away from zero, unlike numpy's bankers' rounding
    rounded = tick_size * consistent_round(
        (price - (diff if prefer_round_down else -diff)) / tick_size
    )
    if zp_math.tolerant_equals(rounded, 0.0):
        return 0.0
    return rounded


def check_stoplimit_prices(price: float, label: str):
    """
    Check to make sure the stop/limit prices are reasonable and raise
    a BadOrderParameters exception if not.
    """
    try:
        if not isfinite(float(price)):
            raise BadOrderParameters(
                msg=f"Attempted to place an order with a {label} price "
                    f"of {price}."
            )
    # This catches arbitrary objects
    except TypeError as exc:
        raise BadOrderParameters(
            msg=f"Attempted to place an order with a {label} price "
                f"of {type(price)}."
        ) from exc

    if price < 0:
        raise BadOrderParameters(
            msg=f"Can't place a {label} order with a negative price."
        )



from ziplime.finance.controls.trading_control import TradingControl


class RestrictedListOrder(TradingControl):
    """TradingControl representing a restricted list of assets that
    cannot be ordered by the algorithm.

    Parameters
    ----------
    restrictions : ziplime.finance.asset_restrictions.Restrictions
        Object representing restrictions of a group of assets.
    """

    def __init__(self, on_error, restrictions):
        super(RestrictedListOrder, self).__init__(on_error)
        self.restrictions = restrictions

    def validate(self, asset, amount, portfolio, algo_datetime, algo_current_data):
        """
        Fail if the asset is in the restricted_list.
        """
        if self.restrictions.is_restricted(asset, algo_datetime):
            self.handle_violation(asset, amount, algo_datetime)



from ziplime.utils.numpy_utils import (
    bool_dtype,
    datetime64ns_dtype,
    float64_dtype,
    int64_dtype,
    object_dtype,
)

CLASSIFIER_DTYPES = frozenset({object_dtype, int64_dtype})
FACTOR_DTYPES = frozenset({datetime64ns_dtype, float64_dtype, int64_dtype})
FILTER_DTYPES = frozenset({bool_dtype})



"""
An ndarray subclass for working with arrays of strings.
"""

from functools import partial, total_ordering
from operator import eq, ne
import re

import numpy as np
from numpy import ndarray
import pandas as pd
from toolz import compose

from ziplime.utils.compat import unicode
from ziplime.utils.functional import instance
from ziplime.utils.numpy_utils import (
    bool_dtype,
    unsigned_int_dtype_with_size_in_bytes,
    is_object,
    object_dtype,
)
from ziplime.utils.pandas_utils import ignore_pandas_nan_categorical_warning

from ziplime.lib.factorize import (
    factorize_strings,
    factorize_strings_known_categories,
    smallest_uint_that_can_hold,
)


def compare_arrays(left, right):
    "Eq check with a short-circuit for identical objects."
    return left is right or ((left.shape == right.shape) and (left == right).all())


def _make_unsupported_method(name):
    def method(*args, **kwargs):
        raise NotImplementedError("Method %s is not supported on LabelArrays." % name)

    method.__name__ = name
    method.__doc__ = "Unsupported LabelArray Method: %s" % name
    return method


class MissingValueMismatch(ValueError):
    """
    Error raised on attempt to perform operations between LabelArrays with
    mismatched missing_values.
    """

    def __init__(self, left, right):
        super(MissingValueMismatch, self).__init__(
            "LabelArray missing_values don't match:"
            " left={}, right={}".format(left, right)
        )


class CategoryMismatch(ValueError):
    """
    Error raised on attempt to perform operations between LabelArrays with
    mismatched category arrays.
    """

    def __init__(self, left, right):
        (mismatches,) = np.where(left != right)
        assert len(mismatches), "Not actually a mismatch!"
        super(CategoryMismatch, self).__init__(
            "LabelArray categories don't match:\n"
            "Mismatched Indices: {mismatches}\n"
            "Left: {left}\n"
            "Right: {right}".format(
                mismatches=mismatches,
                left=left[mismatches],
                right=right[mismatches],
            )
        )


_NotPassed = "_NotPassed"


class LabelArray(ndarray):
    """
    An ndarray subclass for working with arrays of strings.

    Factorizes the input array into integers, but overloads equality on strings
    to check against the factor label.

    Parameters
    ----------
    values : array-like
        Array of values that can be passed to np.asarray with dtype=object.
    missing_value : str
        Scalar value to treat as 'missing' for operations on ``self``.
    categories : list[str], optional
        List of values to use as categories.  If not supplied, categories will
        be inferred as the unique set of entries in ``values``.
    sort : bool, optional
        Whether to sort categories.  If sort is False and categories is
        supplied, they are left in the order provided.  If sort is False and
        categories is None, categories will be constructed in a random order.

    Attributes
    ----------
    categories : ndarray[str]
        An array containing the unique labels of self.
    reverse_categories : dict[str -> int]
        Reverse lookup table for ``categories``. Stores the index in
        ``categories`` at which each entry each unique entry is found.
    missing_value : str or None
        A sentinel missing value with NaN semantics for comparisons.

    Notes
    -----
    Consumers should be cautious when passing instances of LabelArray to numpy
    functions.  We attempt to disallow as many meaningless operations as
    possible, but since a LabelArray is just an ndarray of ints with some
    additional metadata, many numpy functions (for example, trigonometric) will
    happily accept a LabelArray and treat its values as though they were
    integers.

    In a future change, we may be able to disallow more numerical operations by
    creating a wrapper dtype which doesn't register an implementation for most
    numpy ufuncs. Until that change is made, consumers of LabelArray should
    assume that it is undefined behavior to pass a LabelArray to any numpy
    ufunc that operates on semantically-numerical data.

    See Also
    --------
    https://docs.scipy.org/doc/numpy-1.11.0/user/basics.subclassing.html
    """

    SUPPORTED_SCALAR_TYPES = (bytes, unicode, type(None))
    SUPPORTED_NON_NONE_SCALAR_TYPES = (bytes, unicode)

    # @preprocess(
    #     values=coerce(list, partial(np.asarray, dtype=object)),
    #     # Coerce ``list`` to ``list`` to make a copy. Code internally may call
    #     # ``categories.insert(0, missing_value)`` which will mutate this list
    #     # in place.
    #     categories=coerce((list, np.ndarray, set), list),
    # )
    def __new__(cls, values: np.ndarray, missing_value: SUPPORTED_SCALAR_TYPES, categories: list|None=None, sort=True):

        # Numpy's fixed-width string types aren't very efficient. Working with
        # object arrays is faster than bytes or unicode arrays in almost all
        # cases.
        if not is_object(values):
            values = values.astype(object)

        if values.flags.f_contiguous:
            ravel_order = "F"
        else:
            ravel_order = "C"

        if categories is None:
            codes, categories, reverse_categories = factorize_strings(
                values.ravel(ravel_order),
                missing_value=missing_value,
                sort=sort,
            )
        else:
            (
                codes,
                categories,
                reverse_categories,
            ) = factorize_strings_known_categories(
                values.ravel(ravel_order),
                categories=categories,
                missing_value=missing_value,
                sort=sort,
            )
        categories.setflags(write=False)

        return cls.from_codes_and_metadata(
            codes=codes.reshape(values.shape, order=ravel_order),
            categories=categories,
            reverse_categories=reverse_categories,
            missing_value=missing_value,
        )

    @classmethod
    def from_codes_and_metadata(
        cls, codes, categories, reverse_categories, missing_value
    ):
        """
        Rehydrate a LabelArray from the codes and metadata.

        Parameters
        ----------
        codes : np.ndarray[integral]
            The codes for the label array.
        categories : np.ndarray[object]
            The unique string categories.
        reverse_categories : dict[str, int]
            The mapping from category to its code-index.
        missing_value : any
            The value used to represent missing data.
        """
        ret = codes.view(type=cls, dtype=np.void)
        ret._categories = categories
        ret._reverse_categories = reverse_categories
        ret._missing_value = missing_value
        return ret

    @classmethod
    def from_categorical(cls, categorical, missing_value=None):
        """
        Create a LabelArray from a pandas categorical.

        Parameters
        ----------
        categorical : pd.Categorical
            The categorical object to convert.
        missing_value : bytes, unicode, or None, optional
            The missing value to use for this LabelArray.

        Returns
        -------
        la : LabelArray
            The LabelArray representation of this categorical.
        """
        return LabelArray(
            categorical,
            missing_value,
            categorical.categories,
        )

    @property
    def categories(self):
        # This is a property because it should be immutable.
        return self._categories

    @property
    def reverse_categories(self):
        # This is a property because it should be immutable.
        return self._reverse_categories

    @property
    def missing_value(self):
        # This is a property because it should be immutable.
        return self._missing_value

    @property
    def missing_value_code(self):
        return self.reverse_categories[self.missing_value]

    def has_label(self, value):
        return value in self.reverse_categories

    def __array_finalize__(self, obj):
        """
        Called by Numpy after array construction.

        There are three cases where this can happen:

        1. Someone tries to directly construct a new array by doing::

            >>> ndarray.__new__(LabelArray, ...)  # doctest: +SKIP

           In this case, obj will be None.  We treat this as an error case and
           fail.

        2. Someone (most likely our own __new__) does::

           >>> other_array.view(type=LabelArray)  # doctest: +SKIP

           In this case, `self` will be the new LabelArray instance, and
           ``obj` will be the array on which ``view`` is being called.

           The caller of ``obj.view`` is responsible for setting category
           metadata on ``self`` after we exit.

        3. Someone creates a new LabelArray by slicing an existing one.

           In this case, ``obj`` will be the original LabelArray.  We're
           responsible for copying over the parent array's category metadata.
        """
        if obj is None:
            raise TypeError("Direct construction of LabelArrays is not supported.")

        # See docstring for an explanation of when these will or will not be
        # set.
        self._categories = getattr(obj, "categories", None)
        self._reverse_categories = getattr(obj, "reverse_categories", None)
        self._missing_value = getattr(obj, "missing_value", None)

    def as_int_array(self):
        """
        Convert self into a regular ndarray of ints.

        This is an O(1) operation. It does not copy the underlying data.
        """
        return self.view(
            type=ndarray,
            dtype=unsigned_int_dtype_with_size_in_bytes(self.itemsize),
        )

    def as_string_array(self):
        """
        Convert self back into an array of strings.

        This is an O(N) operation.
        """
        return self.categories[self.as_int_array()]

    def as_categorical(self):
        """
        Coerce self into a pandas categorical.

        This is only defined on 1D arrays, since that's all pandas supports.
        """
        if len(self.shape) > 1:
            raise ValueError("Can't convert a 2D array to a categorical.")

        with ignore_pandas_nan_categorical_warning():
            return pd.Categorical.from_codes(
                self.as_int_array(),
                # We need to make a copy because pandas >= 0.17 fails if this
                # buffer isn't writeable.
                self.categories.copy(),
                ordered=False,
            )

    def as_categorical_frame(self, index, columns, name=None):
        """
        Coerce self into a pandas DataFrame of Categoricals.
        """
        if len(self.shape) != 2:
            raise ValueError("Can't convert a non-2D LabelArray into a DataFrame.")

        expected_shape = (len(index), len(columns))
        if expected_shape != self.shape:
            raise ValueError(
                "Can't construct a DataFrame with provided indices:\n\n"
                "LabelArray shape is {actual}, but index and columns imply "
                "that shape should be {expected}.".format(
                    actual=self.shape,
                    expected=expected_shape,
                )
            )

        return pd.Series(
            index=pd.MultiIndex.from_product([index, columns]),
            data=self.ravel().as_categorical(),
            name=name,
        ).unstack()

    def __setitem__(self, indexer, value):
        self_categories = self.categories

        if isinstance(value, self.SUPPORTED_SCALAR_TYPES):
            value_code = self.reverse_categories.get(value, None)
            if value_code is None:
                raise ValueError("%r is not in LabelArray categories." % value)
            self.as_int_array()[indexer] = value_code
        elif isinstance(value, LabelArray):
            value_categories = value.categories
            if compare_arrays(self_categories, value_categories):
                return super(LabelArray, self).__setitem__(indexer, value)
            elif self.missing_value == value.missing_value and set(
                value.categories
            ) <= set(self.categories):
                rhs = LabelArray.from_codes_and_metadata(
                    *factorize_strings_known_categories(
                        value.as_string_array().ravel(),
                        list(self.categories),
                        self.missing_value,
                        False,
                    ),
                    missing_value=self.missing_value,
                ).reshape(value.shape)
                super(LabelArray, self).__setitem__(indexer, rhs)
            else:
                raise CategoryMismatch(self_categories, value_categories)
        else:
            raise NotImplementedError(
                "Setting into a LabelArray with a value of "
                "type {type} is not yet supported.".format(
                    type=type(value).__name__,
                ),
            )

    def set_scalar(self, indexer, value):
        """
        Set scalar value into the array.

        Parameters
        ----------
        indexer : any
            The indexer to set the value at.
        value : str
            The value to assign at the given locations.

        Raises
        ------
        ValueError
            Raised when ``value`` is not a value element of this this label
            array.
        """
        try:
            value_code = self.reverse_categories[value]
        except KeyError as exc:
            raise ValueError("%r is not in LabelArray categories." % value) from exc

        self.as_int_array()[indexer] = value_code

    def __getitem__(self, indexer):
        result = super(LabelArray, self).__getitem__(indexer)
        if result.ndim:
            # Result is still a LabelArray, so we can just return it.
            return result

        # Result is a scalar value, which will be an instance of np.void.
        # Map it back to one of our category entries.
        index = result.view(
            unsigned_int_dtype_with_size_in_bytes(self.itemsize),
        )
        return self.categories[index]

    def is_missing(self):
        """
        Like isnan, but checks for locations where we store missing values.
        """
        return self.as_int_array() == self.reverse_categories[self.missing_value]

    def not_missing(self):
        """
        Like ~isnan, but checks for locations where we store missing values.
        """
        return self.as_int_array() != self.reverse_categories[self.missing_value]

    def _equality_check(op):
        """
        Shared code for __eq__ and __ne__, parameterized on the actual
        comparison operator to use.
        """

        def method(self, other):

            if isinstance(other, LabelArray):
                self_mv = self.missing_value
                other_mv = other.missing_value
                if self_mv != other_mv:
                    raise MissingValueMismatch(self_mv, other_mv)

                self_categories = self.categories
                other_categories = other.categories
                if not compare_arrays(self_categories, other_categories):
                    raise CategoryMismatch(self_categories, other_categories)

                return (
                    op(self.as_int_array(), other.as_int_array())
                    & self.not_missing()
                    & other.not_missing()
                )

            elif isinstance(other, ndarray):
                # Compare to ndarrays as though we were an array of strings.
                # This is fairly expensive, and should generally be avoided.
                return op(self.as_string_array(), other) & self.not_missing()

            elif isinstance(other, self.SUPPORTED_SCALAR_TYPES):
                i = self._reverse_categories.get(other, -1)
                return op(self.as_int_array(), i) & self.not_missing()

            return op(super(LabelArray, self), other)

        return method

    __eq__ = _equality_check(eq)
    __ne__ = _equality_check(ne)
    del _equality_check

    def view(self, dtype=_NotPassed, type=_NotPassed):
        if type is _NotPassed and dtype not in (_NotPassed, self.dtype):
            raise TypeError("Can't view LabelArray as another dtype.")

        # The text signature on ndarray.view makes it look like the default
        # values for dtype and type are `None`, but passing None explicitly has
        # different semantics than not passing an arg at all, so we reconstruct
        # the kwargs dict here to simulate the args not being passed at all.
        kwargs = {}
        if dtype is not _NotPassed:
            kwargs["dtype"] = dtype
        if type is not _NotPassed:
            kwargs["type"] = type
        return super(LabelArray, self).view(**kwargs)

    def astype(self, dtype, order="K", casting="unsafe", subok=True, copy=True):
        if dtype == self.dtype:
            if not subok:
                array = self.view(type=np.ndarray)
            else:
                array = self

            if copy:
                return array.copy()
            return array

        if dtype == object_dtype:
            return self.as_string_array()

        if dtype.kind == "S":
            return self.as_string_array().astype(
                dtype,
                order=order,
                casting=casting,
                subok=subok,
                copy=copy,
            )

        raise TypeError(
            "%s can only be converted into object, string, or void,"
            " got: %r"
            % (
                type(self).__name__,
                dtype,
            ),
        )

    # In general, we support resizing, slicing, and reshaping methods, but not
    # numeric methods.
    SUPPORTED_NDARRAY_METHODS = frozenset(
        [
            "astype",
            "base",
            "compress",
            "copy",
            "data",
            "diagonal",
            "dtype",
            "flat",
            "flatten",
            "item",
            "itemset",
            "itemsize",
            "nbytes",
            "ndim",
            "ravel",
            "repeat",
            "reshape",
            "resize",
            "setflags",
            "shape",
            "size",
            "squeeze",
            "strides",
            "swapaxes",
            "take",
            "trace",
            "transpose",
            "view",
        ]
    )
    PUBLIC_NDARRAY_METHODS = frozenset(
        [s for s in dir(ndarray) if not s.startswith("_")]
    )

    # Generate failing wrappers for all unsupported methods.
    locals().update(
        {
            method: _make_unsupported_method(method)
            for method in PUBLIC_NDARRAY_METHODS - SUPPORTED_NDARRAY_METHODS
        }
    )

    def __repr__(self):
        repr_lines = repr(self.as_string_array()).splitlines()
        repr_lines[0] = repr_lines[0].replace("array(", "LabelArray(", 1)
        repr_lines[-1] = repr_lines[-1].rsplit(",", 1)[0] + ")"
        # The extra spaces here account for the difference in length between
        # 'array(' and 'LabelArray('.
        return "\n     ".join(repr_lines)

    def empty_like(self, shape):
        """
        Make an empty LabelArray with the same categories as ``self``, filled
        with ``self.missing_value``.
        """
        return type(self).from_codes_and_metadata(
            codes=np.full(
                shape,
                self.reverse_categories[self.missing_value],
                dtype=unsigned_int_dtype_with_size_in_bytes(self.itemsize),
            ),
            categories=self.categories,
            reverse_categories=self.reverse_categories,
            missing_value=self.missing_value,
        )

    def map_predicate(self, f):
        """
        Map a function from str -> bool element-wise over ``self``.

        ``f`` will be applied exactly once to each non-missing unique value in
        ``self``. Missing values will always return False.
        """
        # Functions passed to this are of type str -> bool.  Don't ever call
        # them on None, which is the only non-str value we ever store in
        # categories.
        if self.missing_value is None:

            def f_to_use(x):
                return False if x is None else f(x)

        else:
            f_to_use = f

        # Call f on each unique value in our categories.
        results = np.vectorize(f_to_use, otypes=[bool_dtype])(self.categories)

        # missing_value should produce False no matter what
        results[self.reverse_categories[self.missing_value]] = False

        # unpack the results form each unique value into their corresponding
        # locations in our indices.
        return results[self.as_int_array()]

    def map(self, f):
        """
        Map a function from str -> str element-wise over ``self``.

        ``f`` will be applied exactly once to each non-missing unique value in
        ``self``. Missing values will always map to ``self.missing_value``.
        """
        # f() should only return None if None is our missing value.
        if self.missing_value is None:
            allowed_outtypes = self.SUPPORTED_SCALAR_TYPES
        else:
            allowed_outtypes = self.SUPPORTED_NON_NONE_SCALAR_TYPES

        def f_to_use(x, missing_value=self.missing_value, otypes=allowed_outtypes):

            # Don't call f on the missing value; those locations don't exist
            # semantically. We return _sortable_sentinel rather than None
            # because the np.unique call below sorts the categories array,
            # which raises an error on Python 3 because None and str aren't
            # comparable.
            if x == missing_value:
                return _sortable_sentinel

            ret = f(x)

            if not isinstance(ret, otypes):
                raise TypeError(
                    "LabelArray.map expected function {f} to return a string"
                    " or None, but got {type} instead.\n"
                    "Value was {value}.".format(
                        f=f.__name__,
                        type=type(ret).__name__,
                        value=ret,
                    )
                )

            if ret == missing_value:
                return _sortable_sentinel

            return ret

        new_categories_with_duplicates = np.vectorize(f_to_use, otypes=[object])(
            self.categories
        )

        # If f() maps multiple inputs to the same output, then we can end up
        # with the same code duplicated multiple times. Compress the categories
        # by running them through np.unique, and then use the reverse lookup
        # table to compress codes as well.
        new_categories, bloated_inverse_index = np.unique(
            new_categories_with_duplicates, return_inverse=True
        )

        if new_categories[0] is _sortable_sentinel:
            # f_to_use return _sortable_sentinel for locations that should be
            # missing values in our output. Since np.unique returns the uniques
            # in sorted order, and since _sortable_sentinel sorts before any
            # string, we only need to check the first array entry.
            new_categories[0] = self.missing_value

        # `reverse_index` will always be a 64 bit integer even if we can hold a
        # smaller array.
        reverse_index = bloated_inverse_index.astype(
            smallest_uint_that_can_hold(len(new_categories))
        )
        new_codes = np.take(reverse_index, self.as_int_array())

        return self.from_codes_and_metadata(
            new_codes,
            new_categories,
            dict(zip(new_categories, range(len(new_categories)))),
            missing_value=self.missing_value,
        )

    def startswith(self, prefix):
        """
        Element-wise startswith.

        Parameters
        ----------
        prefix : str

        Returns
        -------
        matches : np.ndarray[bool]
            An array with the same shape as self indicating whether each
            element of self started with ``prefix``.
        """
        return self.map_predicate(lambda elem: elem.startswith(prefix))

    def endswith(self, suffix):
        """
        Elementwise endswith.

        Parameters
        ----------
        suffix : str

        Returns
        -------
        matches : np.ndarray[bool]
            An array with the same shape as self indicating whether each
            element of self ended with ``suffix``
        """
        return self.map_predicate(lambda elem: elem.endswith(suffix))

    def has_substring(self, substring):
        """
        Elementwise contains.

        Parameters
        ----------
        substring : str

        Returns
        -------
        matches : np.ndarray[bool]
            An array with the same shape as self indicating whether each
            element of self ended with ``suffix``.
        """
        return self.map_predicate(lambda elem: substring in elem)

    # @preprocess(pattern=coerce(from_=(bytes, unicode), to=re.compile))
    def matches(self, pattern: re.Pattern):
        """
        Elementwise regex match.

        Parameters
        ----------
        pattern : str or compiled regex

        Returns
        -------
        matches : np.ndarray[bool]
            An array with the same shape as self indicating whether each
            element of self was matched by ``pattern``.
        """
        return self.map_predicate(compose(bool, pattern.match))

    # These types all implement an O(N) __contains__, so pre-emptively
    # coerce to `set`.
    #@preprocess(container=coerce((list, tuple, np.ndarray), set))
    def element_of(self, container:set):
        """
        Check if each element of self is an of ``container``.

        Parameters
        ----------
        container : object
            An object implementing a __contains__ to call on each element of
            ``self``.

        Returns
        -------
        is_contained : np.ndarray[bool]
            An array with the same shape as self indicating whether each
            element of self was an element of ``container``.
        """
        return self.map_predicate(container.__contains__)


@instance  # This makes _sortable_sentinel a singleton instance.
@total_ordering
class _sortable_sentinel:
    """Dummy object that sorts before any other python object."""

    def __eq__(self, other):
        return self is other

    def __lt__(self, other):
        return True


def labelarray_where(cond, trues: LabelArray, falses: LabelArray):
    """LabelArray-aware implementation of np.where."""
    if trues.missing_value != falses.missing_value:
        raise ValueError("Can't compute where on arrays with different missing values.")

    strs = np.where(cond, trues.as_string_array(), falses.as_string_array())
    return LabelArray(strs, missing_value=trues.missing_value)



import datetime
from dataclasses import dataclass

@dataclass
class BundleMetadata:
    name: str
    version: str
    calendar_name: str
    start_date: datetime.datetime
    end_date: datetime.datetime
    frequency: datetime.timedelta



import datetime
from typing import Any

import pandas as pd
from exchange_calendars import ExchangeCalendar


from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger
from ziplime.sources.benchmark_source import BenchmarkSource




class CashFlow:
    """Tracks daily and cumulative cash flow.

    Notes
    -----
    For historical reasons, this field is named 'capital_used' in the packets.
    """

    def start_of_simulation(
            self, ledger: Ledger, emission_rate: datetime.timedelta, trading_calendar: ExchangeCalendar,
            sessions: pd.DatetimeIndex, benchmark_source: BenchmarkSource
    ):
        self._previous_cash_flow = 0.0

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        cash_flow = ledger.portfolio.cash_flow
        packet["minute_perf"]["capital_used"] = cash_flow - self._previous_cash_flow
        packet["cumulative_perf"]["capital_used"] = cash_flow

    def end_of_session(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                       exchanges: dict[str, Exchange]):
        cash_flow = ledger.portfolio.cash_flow
        packet["daily_perf"]["capital_used"] = cash_flow - self._previous_cash_flow
        packet["cumulative_perf"]["capital_used"] = cash_flow
        self._previous_cash_flow = cash_flow
